/*
 * This file is part of Siril, an astronomy image processor.
 * Copyright (C) 2005-2011 Francois Meyer (dulle at free.fr)
 * Copyright (C) 2012-2014 team free-astro (see more in AUTHORS file)
 * Reference site is http://free-astro.vinvin.tf/index.php/Siril
 *
 * Siril is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Siril is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Siril. If not, see <http://www.gnu.org/licenses/>.
*/

#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/time.h>
#include <dirent.h>
#include <limits.h>
#include <stdarg.h>
#include <math.h>	// for M_PI
#include <libgen.h>

#include "siril.h"
#include "conversion.h"
#include "callbacks.h"
#include "proto.h"
#include "colors.h"
#include "films.h"
#include "command.h"	// for processcommand()
#include "PSF.h"
#include "PSF_list.h"
#include "star_finder.h"
#include "histogram.h"
#include "single_image.h"
#include "gradient.h"
#include "registration.h"
#include "stacking.h"
#include "fft.h"
#include "compositing.h"
#ifdef HAVE_OPENCV
#include "opencv.h"
#endif
#include "Def_Wavelet.h"

static enum { CD_NULL, CD_INCALL, CD_EXCALL } confirm;

layer_info predefined_layers_colors[] = {
	/* name, lambda, lo, hi, c/over, c/under, mode */
	{ "Luminance", 0.0, 0, 0, FALSE, FALSE, NORMAL_DISPLAY },// no color, undefined value is <0
	{ "Red", 650.0, 0, 0, FALSE, FALSE, NORMAL_DISPLAY },	// approx. of the middle of the color
	{ "Green", 530.0, 0, 0, FALSE, FALSE, NORMAL_DISPLAY },	// approx. of the middle of the color
	{ "Blue", 450.0, 0, 0, FALSE, FALSE, NORMAL_DISPLAY }	// approx. of the middle of the color
};

/* remap index data, an index for each layer */
static BYTE *remap_index[MAXGRAYVPORT];
static float last_pente[MAXGRAYVPORT];
static display_mode last_mode[MAXGRAYVPORT];

GtkWidget* lookup_widget (const gchar *widget_name) {
	return GTK_WIDGET(gtk_builder_get_object(builder, widget_name));
}

static gboolean idle_messaging(gpointer p);

void initialize_remap() {
	int i;
	for (i=0; i<MAXGRAYVPORT; i++) {
		remap_index[i] = NULL;
		last_pente[i] = 0.f;
		last_mode[i] = HISTEQ_DISPLAY;
		// only HISTEQ mode always computes the index, it's a good initializer here
	}
}

void on_register_all_toggle(GtkToggleButton *togglebutton, gpointer user_data) {
	update_reg_interface(TRUE);
}

struct log_message {
	char *timestamp;
	char *message;
	const char* color;
};

/* This function writes a message on Siril's console/log. It is not thread safe.
 * There is a limit in number of characters that it is able to write in one call: 1023.
 * Return value is the string printed from arguments, or NULL if argument was empty or
 * only newline. It is an allocated string and must not be freed. It can be
 * reused until next call to this function.
 */
char* siril_log_internal(const char* format, const char* color, va_list arglist) {
	static char *msg = NULL;
	struct tm *now;
	time_t now_sec;
	char timestamp[30];
	struct log_message *new_msg;

	if (msg == NULL) {
		msg = malloc(1024);
		msg[1023] = '\0';
	}

	vsnprintf(msg, 1023, format, arglist);

	if(msg == NULL || msg[0] == '\0')
		return NULL;

	if (msg[0] == '\n' && msg[1] == '\0') {
		fputc('\n', stdout);
		new_msg = malloc(sizeof(struct log_message));
		new_msg->timestamp = NULL;
		new_msg->message = "\n";
		new_msg->color = NULL;
		gdk_threads_add_idle(idle_messaging, new_msg);
		return NULL;
	}

	fprintf(stdout, "log: %s", msg);
	now_sec = time(NULL);
	now = localtime(&now_sec);
	sprintf(timestamp, "%.2d:%.2d:%.2d: ", now->tm_hour, now->tm_min, now->tm_sec);

	new_msg = malloc(sizeof(struct log_message));
	new_msg->timestamp = strdup(timestamp);
	new_msg->message = strdup(msg);
	new_msg->color = color;
	gdk_threads_add_idle(idle_messaging, new_msg);

	return msg;
}

// The main thread internal function that does the printing.
static gboolean idle_messaging(gpointer p) {
	static GtkTextBuffer *tbuf = NULL;
	static GtkTextView *text = NULL;
	GtkTextIter iter;
	struct log_message *log = (struct log_message *)p;

	if (!tbuf) {
		text = GTK_TEXT_VIEW(gtk_builder_get_object(builder, "output"));
		tbuf = gtk_text_view_get_buffer(text);
	}

	if (log->message[0] == '\n' && log->message[1] == '\0') {
		gtk_text_buffer_get_start_iter(tbuf, &iter);
		gtk_text_buffer_insert(tbuf, &iter, log->message, strlen(log->message));
		free(log);
		return FALSE;
	}

	PangoFontDescription *font = pango_font_description_from_string( "Mono" );
	gtk_widget_override_font(GTK_WIDGET(text), font );

	gtk_text_buffer_get_end_iter(tbuf, &iter);
	gtk_text_buffer_insert_with_tags_by_name(tbuf, &iter, log->timestamp, strlen(log->timestamp), "bold", NULL);

	if (log->color == NULL)
		gtk_text_buffer_insert_with_tags_by_name(tbuf, &iter,
				log->message, strlen(log->message), "normal", NULL);
	else gtk_text_buffer_insert_with_tags_by_name(tbuf, &iter,
			log->message, strlen(log->message), log->color, NULL);

	pango_font_description_free(font);

	/* scroll to end */
	gtk_text_buffer_get_end_iter (tbuf, &iter);
	GtkTextMark *insert_mark = gtk_text_buffer_get_insert (tbuf);
	gtk_text_buffer_place_cursor(tbuf, &iter);
	gtk_text_view_scroll_to_mark (text, insert_mark, 0.0, TRUE, 0.0, 1.0);
	gtk_widget_queue_draw(GTK_WIDGET(text));
	
	free(log->timestamp);
	free(log->message);
	free(log);
	return FALSE;
}

char* siril_log_message(const char* format, ...) {
	va_list args;
	va_start(args, format);
	g_mutex_lock(&com.mutex);
	char *msg = siril_log_internal(format, NULL, args);
	g_mutex_unlock(&com.mutex);
	va_end(args);
	return msg;
}

char* siril_log_color_message(const char* format, const char* color, ...) {
	va_list args;
	va_start(args, color);
	g_mutex_lock(&com.mutex);
	char *msg = siril_log_internal(format, color, args);
	g_mutex_unlock(&com.mutex);
	va_end(args);
	return msg;
}

void show_time (struct timeval t_start, struct timeval t_end) {
	float time = (float)(((t_end.tv_sec - t_start.tv_sec)*1000000L + t_end.tv_usec) - t_start.tv_usec)/1000000.0;
	
	if (time > 60.f) {
		int min = (int)time / 60;
		int sec = (int)time % 60 + 1;
		siril_log_color_message("Execution time: %d min %.2d s.\n", "green", min, sec);
	}
	else if (time < 1.f) {
		float ms = time * 1000.f;
		siril_log_color_message("Execution time: %.2f ms.\n", "green", ms);
	}	
	else {
		siril_log_color_message("Execution time: %.2f s.\n", "green", time);
	}
}

gboolean inimage(GdkEvent *event){
	double zoom = get_zoom_val();
	return ((GdkEventButton*)event)->x > 0 &&
		((GdkEventButton*)event)->x < gfit.rx * zoom &&
		((GdkEventButton*)event)->y > 0 &&
		((GdkEventButton*)event)->y < gfit.ry * zoom;
}

/* Sets maximum value for contrast scales. Minimum is always 0.
 * Should be done on first image load of a sequence and when single images are loaded.
 * Max value is taken from gfit.maxi, recomputed if not present.
 *
 * TODO: here we have USHRT_MAX as upper value for the sliders, should we have
 * gfit.max[layer] instead? If yes, this function should be called on every layer tab change.
 */
void set_cutoff_sliders_max_values() {
	static GtkAdjustment *adj1 = NULL, *adj2 = NULL;
	gdouble max_val;
	if (adj1 == NULL) {
		adj1 = GTK_ADJUSTMENT(gtk_builder_get_object(builder, "adjustment1"));	// scalemax
		adj2 = GTK_ADJUSTMENT(gtk_builder_get_object(builder, "adjustment2"));	// scalemin
	}
	fprintf(stdout, "Setting MAX value for cutoff sliders adjustments\n");
	/* set max value for range according to number of bits of original image
	 * We should use gfit.bitpix for this, but it's currently always USHORT_IMG */
	/*if (gfit.bitpix == BYTE_IMG)
		max_val = UCHAR_MAX_DOUBLE;
	else max_val = USHRT_MAX_DOUBLE;*/
	if (gfit.maxi == 0)
		image_find_minmax(&gfit, 0);
	if (gfit.maxi <= UCHAR_MAX)
		max_val = UCHAR_MAX_DOUBLE;
	else max_val = USHRT_MAX_DOUBLE;

	gtk_adjustment_set_upper(adj1, max_val);
	gtk_adjustment_set_upper(adj2, max_val);
}

/* Sets the value scalemin and scalemax sliders so that they match the hi and
 * lo values stored for the current viewport.
 * It also sets the 'cut' checkbox status.
 * The function should be called on every tab change and file opening */
void set_cutoff_sliders_values() {
	gchar buffer[10];
	static GtkAdjustment *adjmin = NULL, *adjmax = NULL;
	static GtkEntry *maxentry = NULL, *minentry = NULL;
	static GtkToggleButton *cutmax = NULL;
	WORD hi, lo;
	gboolean cut_over;
	if (adjmin == NULL) {
		adjmax = GTK_ADJUSTMENT(gtk_builder_get_object(builder, "adjustment1")); // scalemax
		adjmin = GTK_ADJUSTMENT(gtk_builder_get_object(builder, "adjustment2")); // scalemin
		maxentry = GTK_ENTRY(gtk_builder_get_object(builder, "max_entry"));
		minentry = GTK_ENTRY(gtk_builder_get_object(builder, "min_entry"));
		cutmax = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "checkcut_max"));
	}
	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers && com.uniq->layers) {
		hi = com.uniq->layers[com.cvport].hi;
		lo = com.uniq->layers[com.cvport].lo;
		cut_over = com.uniq->layers[com.cvport].cut_over;
	}
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers && com.seq.layers) {
		hi = com.seq.layers[com.cvport].hi;
		lo = com.seq.layers[com.cvport].lo;
		cut_over = com.seq.layers[com.cvport].cut_over;
	}
	else return;	// there should be no other normal cases
	fprintf(stdout, "setting ranges scalemin=%d, scalemax=%d\n", lo, hi);
	gtk_adjustment_set_value (adjmin, (gdouble)lo);
	gtk_adjustment_set_value (adjmax, (gdouble)hi);
	snprintf(buffer, 6, "%u", hi);
	g_signal_handlers_block_by_func(maxentry, on_min_entry_changed, NULL);
	gtk_entry_set_text(maxentry, buffer);
	g_signal_handlers_unblock_by_func(maxentry, on_min_entry_changed, NULL);
	snprintf(buffer, 6, "%u", lo);
	g_signal_handlers_block_by_func(minentry, on_min_entry_changed, NULL);
	gtk_entry_set_text(minentry, buffer);
	g_signal_handlers_unblock_by_func(minentry, on_min_entry_changed, NULL);
	gtk_toggle_button_set_active(cutmax, cut_over);
}

/* Initialize the combobox when loading new single_image */
void initialize_display_mode() {
	static GtkComboBox *modecombo = NULL;
	static GtkToggleButton *chainedbutton = NULL;
	display_mode mode;
	int i, raw_mode;

	if (!modecombo)
		modecombo = GTK_COMBO_BOX(lookup_widget("combodisplay"));
	raw_mode = gtk_combo_box_get_active(modecombo);
	/* Check if never initialized. In this case the mode is set to linear */
	if (raw_mode == -1)
		mode = NORMAL_DISPLAY;
	else mode = raw_mode;
	/* The mode is applyed for each layer */
	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers) {
		for (i=0; i<com.uniq->nb_layers; i++)
			com.uniq->layers[i].rendering_mode = mode;
	}
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers){
		for (i=0; i<com.seq.nb_layers; i++)
			com.seq.layers[i].rendering_mode = mode;
	}
	/* In the case where the layer were unchained, we chaine it */
	if (!chainedbutton)
		chainedbutton = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_chain"));
	if (!gtk_toggle_button_get_active(chainedbutton)) {
		g_signal_handlers_block_by_func(chainedbutton, on_checkchain_toggled, NULL);
		gtk_toggle_button_set_active(chainedbutton, TRUE);
		g_signal_handlers_unblock_by_func(chainedbutton, on_checkchain_toggled, NULL);
	}

}

/* Sets the display mode combo box to the value stored in the relevant struct.
 * The operation is purely graphical. */
void set_display_mode() {
	static GtkComboBox *modecombo = NULL;
	display_mode mode;

	if (!modecombo)
		modecombo = GTK_COMBO_BOX(lookup_widget("combodisplay"));
	
	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers && com.uniq->layers)
		mode = com.uniq->layers[com.cvport].rendering_mode;
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers && com.seq.layers)
		mode = com.seq.layers[com.cvport].rendering_mode;
	else return;

	g_signal_handlers_block_by_func(modecombo, on_combodisplay_changed, NULL);
	gtk_combo_box_set_active(modecombo, mode);
	g_signal_handlers_unblock_by_func(modecombo, on_combodisplay_changed, NULL);
}

/* called when an image is included or excluded from the sequence.
 * it sets the toggle button "exclude_button" so it matches the inclusion state.
 * n is the image number in the sequence
 * changed indicates if a value was changed and if the display needs to be refreshed.
 */
void adjust_exclude(int n, gboolean changed) {
	static GtkWidget *excl_butt = NULL;
	if (!com.seq.imgparam || n<0 || n>=com.seq.number) return;
	if (excl_butt == NULL)
		excl_butt = lookup_widget("exclude_button");

	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(excl_butt)) == 
			com.seq.imgparam[n].incl) {
		g_signal_handlers_block_by_func(excl_butt, on_excludebutton_toggled, NULL);
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(excl_butt), !com.seq.imgparam[n].incl);
		g_signal_handlers_unblock_by_func(excl_butt, on_excludebutton_toggled, NULL);
	}

	if (changed) {
		redraw(com.cvport, REMAP_NONE);
		adjust_sellabel();
	}
}

void adjust_refimage(int n){
	static GtkWidget *ref_butt = NULL;
	if (ref_butt == NULL)
		ref_butt = lookup_widget("refframe");

	//fprintf(stdout, "adjust refimage: %d (ref is %d)\n", n, com.seq.reference_image); 
	g_signal_handlers_block_by_func(ref_butt, on_ref_frame_toggled, NULL);
	if (com.seq.reference_image == n) {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ref_butt), TRUE);
	} else {
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ref_butt), FALSE);
	}
	g_signal_handlers_unblock_by_func(ref_butt, on_ref_frame_toggled, NULL);
}

void draw_empty_image(cairo_t *cr, guint width, guint height) {
	// black image with red square
	cairo_set_source_rgb(cr, 0,0,0);
	cairo_rectangle (cr, 0,0, width, height);
	cairo_fill(cr);
	cairo_set_source_rgb(cr, 0.3,0,0);
	cairo_rectangle (cr, 100, 70, 50, 50);
	cairo_fill(cr);
}

/* callback for GtkDrawingArea, draw event
 * see http://developer.gnome.org/gtk3/3.2/GtkDrawingArea.html
 * http://developer.gnome.org/gdk-pixbuf/stable/gdk-pixbuf-Image-Data-in-Memory.html
 * http://www.cairographics.org/manual/
 * http://www.cairographics.org/manual/cairo-Image-Surfaces.html#cairo-image-surface-create-for-data
 */
gboolean redraw_drawingarea(GtkWidget *widget, cairo_t *cr, gpointer data)
{
	int image_width, image_height, window_width, window_height;
	//fits *fit=&(gfit);
	int vport = -1, i;
	double zoom;

	// we need to identify which vport is being redrawn
	vport = match_drawing_area_widget(widget, TRUE);
	if (vport == -1) {
		fprintf(stderr, "COULD NOT FIND THE VPORT FOR THE DRAW CALLBACK\n");
		return TRUE;
	}
	//fprintf(stdout, "drawing vport %d\n", vport);

	window_width = gtk_widget_get_allocated_width(widget);
	window_height = gtk_widget_get_allocated_height(widget);
	zoom = get_zoom_val();
	image_width = (int)(((double)window_width) / zoom);
	image_height = (int)(((double)window_height) / zoom);
	//gtk_style_context_get_color (gtk_widget_get_style_context(widget), 0, &color);
	//gdk_cairo_set_source_rgba (cr, &color);

	if (vport == RGB_VPORT) {
		if (com.rgbbuf) {
			cairo_scale(cr, zoom, zoom);
			cairo_set_source_surface(cr, com.surface[RGB_VPORT], 0, 0);
			cairo_paint(cr);
		} else {
			fprintf(stdout, "RGB buffer is empty, drawing black image\n");
			draw_empty_image(cr, window_width, window_height);
		}
	} else {
		if (com.graybuf[vport]) {
			cairo_scale(cr, zoom, zoom);
			cairo_set_source_surface(cr, com.surface[vport], 0, 0);
			cairo_paint(cr);
		} else {
			fprintf(stdout, "Buffer %d is empty, drawing black image\n", vport);
			draw_empty_image(cr, window_width, window_height);
		}
	}

	/* draw the selection rectangle */
	if (com.selection.w > 0 && com.selection.h > 0) {
		static double dash_format[] = { 4.0, 2.0 };
		/* fprintf(stdout, "drawing the selection rectangle (%d,%d) (%d,%d)\n",
				com.selection.x, com.selection.y, com.selection.w, com.selection.h); */
		cairo_set_line_width (cr, 0.8/zoom);
		cairo_set_dash(cr, dash_format, 2, 0);
		cairo_set_source_rgb(cr, 0.8, 1.0, 0.8);
		cairo_rectangle(cr, (double)com.selection.x, (double)com.selection.y,
				(double)com.selection.w, (double)com.selection.h);
		cairo_stroke(cr);
	}

	/* draw a cross on excluded images */
	if (sequence_is_loaded() &&
			com.seq.imgparam && com.seq.current >= 0 &&
			!com.seq.imgparam[com.seq.current].incl) {
		if (image_width > gfit.rx) image_width = gfit.rx;
		if (image_height > gfit.ry) image_height = gfit.ry;
		cairo_set_dash(cr, NULL, 0, 0);
		cairo_set_source_rgb(cr, 1.0, 0.8, 0.7);
		cairo_set_line_width (cr, 2.0/zoom);
		cairo_move_to(cr, 0, 0);
		cairo_line_to(cr, image_width, image_height);
		cairo_move_to(cr, 0, image_height);
		cairo_line_to(cr, image_width, 0);
		cairo_stroke(cr);
	}

	/* draw detected stars and highlight the selected star */
	if (com.stars) {
		/* com.stars is a NULL-terminated array */
		int i = 0;
		cairo_set_dash(cr, NULL, 0, 0);
		cairo_set_source_rgba(cr, 1.0, 0.4, 0.0, 0.9);
		cairo_set_line_width(cr, 1.5);
		while (com.stars[i]) {
			//if (com.stars[i]->xpos <= 0.0) { ++i; continue; }
			double size = sqrt(com.stars[i]->Sx/2.)*2*sqrt(log(2.)*2); // by design Sx>Sy. We redefine FWHM to be sure to have the value in px
			if (i == com.selected_star) {
				cairo_set_line_width(cr, 2);
				cairo_set_source_rgba(cr, 0.0, 0.4, 1.0, 0.6);
				cairo_rectangle(cr, com.stars[i]->xpos-1.5*size, com.stars[i]->ypos-1.5*size, 3*size, 3*size);
				cairo_stroke(cr);
				cairo_set_line_width(cr, 1.5);
				cairo_set_source_rgba(cr, 1.0, 0.4, 0.0, 0.9);
			}
			cairo_arc(cr, com.stars[i]->xpos, com.stars[i]->ypos, size, 0., 2. * M_PI);
			cairo_stroke(cr);
			i++;
		}
//		fprintf(stdout, "Drawn %d stars.\n", i);
	}

	/* draw preview rectangles */
	if (sequence_is_loaded()) {
		for (i=0; i<PREVIEW_NB; i++) {
			if (com.seq.previewX[i] >= 0) {
				int textX, textY;
				char text[3];
				cairo_set_line_width (cr, 0.5/zoom);
				cairo_set_source_rgb(cr, 0.1, 0.6, 0.0);
				cairo_rectangle(cr, com.seq.previewX[i]-com.seq.previewW[i]/2,
						com.seq.previewY[i]-com.seq.previewH[i]/2,
						com.seq.previewW[i], com.seq.previewH[i]);
				cairo_stroke(cr);

				textX = com.seq.previewX[i]-com.seq.previewW[i]/2;
				if (textX < 0)
					textX += com.seq.previewW[i]-20;
				else textX += 15;
				textY = com.seq.previewY[i]-com.seq.previewH[i]/2;
				if (textY < 0)
					textY += com.seq.previewH[i]-15;
				else textY += 20;
				sprintf(text, "%d", i+1);

				cairo_set_font_size(cr, 12.0/zoom);
				cairo_move_to(cr, textX, textY);
				cairo_show_text(cr, text);
			}
		}
	}
	
	if (com.grad && com.grad_boxes_drawn) {
		int i = 0;
		while (i < com.grad_nb_boxes) {
			if (com.grad[i].boxvalue != -1.0) {
				cairo_set_line_width(cr, 1.5);
				cairo_set_source_rgba(cr, 0.2, 1.0, 0.3, 1.0);
				cairo_rectangle(cr, com.grad[i].centre.x-com.grad_size_boxes, com.grad[i].centre.y-com.grad_size_boxes,
							 com.grad_size_boxes, com.grad_size_boxes);
				cairo_stroke(cr);
			}
			i++;
		}
	}

	return FALSE;
}

/* fill the label indicating how many images are selected in the sequence and
 * which one is the reference image, at the bottom of the main window */
int adjust_sellabel(){
	static GtkLabel *local_label = NULL, *global_label = NULL;
	char bufferlocal[256], bufferglobal[256], format[150];
	if (local_label == NULL) {
		local_label = GTK_LABEL(lookup_widget("imagesel_label"));
		global_label = GTK_LABEL(lookup_widget("labelseq"));
	}
	if (sequence_is_loaded()) {
		if (com.seq.reference_image != -1) {
			if (com.seq.fixed <= 1){
				sprintf(format, "<%%s.seq> : %%d images selected out of %%d, reference image is %%d");
			} else {
				sprintf(format, "<%%s.seq> : %%d images selected out of %%d, reference image is %%.%dd", com.seq.fixed);
			}
			snprintf(bufferlocal, 255, format,
					com.seq.seqname,
					com.seq.selnum,
					com.seq.number,
					com.seq.imgparam[com.seq.reference_image].filenum);

		} else {
			snprintf(bufferlocal, 255,
					"<%s.seq> : %d images selected out of %d, no reference image set",
					com.seq.seqname,
					com.seq.selnum,
					com.seq.number);
		}
		snprintf(bufferglobal, 255, "%s, %d images selected", com.seq.seqname, com.seq.selnum);
		//gtk_widget_set_sensitive(lookup_widget("goregister_button"), com.seq.selnum>0?TRUE:FALSE);
	} else {
		sprintf(bufferlocal, "No sequence");
		sprintf(bufferglobal, "- none -");
		gtk_widget_set_sensitive(lookup_widget("goregister_button"), FALSE);
	}

	gtk_label_set_text (local_label, bufferlocal);
	gtk_label_set_text (global_label, bufferglobal);
	return 0;
}

struct _label_data {
	const char *label_name;
	char *text;
};

static gboolean set_label_text_idle(gpointer p) {
	struct _label_data *args = (struct _label_data *)p;
	GtkLabel *label = GTK_LABEL(gtk_builder_get_object(builder, args->label_name));
	gtk_label_set_text(label, args->text);
	free(args->text);
	free(args);
	return FALSE;
}

void set_label_text_from_main_thread(const char *label_name, const char *text) {
	struct _label_data *data = malloc(sizeof(struct _label_data));
	data->label_name = label_name;
	data->text = strdup(text);
	gdk_threads_add_idle(set_label_text_idle, data);
}

void set_GUI_CWD() {
	if (!com.wd) return;
	GtkLabel *label = GTK_LABEL(lookup_widget("labelcwd"));
	gtk_label_set_text(label, com.wd);
	//set_label_text_from_main_thread("labelcwd", com.wd);
}

/* size is in kiB */
void set_GUI_MEM(unsigned long size) {
	char str[20];
	sprintf(str, "Mem: %ldMB", size/1024);
	set_label_text_from_main_thread("labelmem", str);
}

void remaprgb(void){
	guchar *dst;
	guchar *bufr, *bufg, *bufb;
	gint i, j;
	int nbdata;

	fprintf(stderr, "remaprgb\n");
	if(!isrgb(&gfit))
		return;

	// allocate if not already done or the same size
	if (cairo_format_stride_for_width(CAIRO_FORMAT_RGB24, gfit.rx) !=
			com.surface_stride[RGB_VPORT] ||
			gfit.ry != com.surface_height[RGB_VPORT] ||
			!com.surface[RGB_VPORT] || !com.rgbbuf) {
		guchar *oldbuf = com.rgbbuf;
		fprintf(stderr, "RGB display buffers and surface (re-)allocation\n");
		com.surface_stride[RGB_VPORT] = cairo_format_stride_for_width(CAIRO_FORMAT_RGB24, gfit.rx);
		com.surface_height[RGB_VPORT] = gfit.ry;
		com.rgbbuf = realloc(com.rgbbuf,
				com.surface_stride[RGB_VPORT] * gfit.ry * sizeof(guchar));
		if (com.rgbbuf == NULL) {
			fprintf(stderr, "Could not allocate memory for RGB buffer (out of memory?)\n");
			if (oldbuf) free(oldbuf);
			return;
		}
		if (com.surface[RGB_VPORT])
			cairo_surface_destroy(com.surface[RGB_VPORT]);
		com.surface[RGB_VPORT] = cairo_image_surface_create_for_data(com.rgbbuf,
				CAIRO_FORMAT_RGB24, gfit.rx, gfit.ry,
				com.surface_stride[RGB_VPORT]);
		if (cairo_surface_status(com.surface[RGB_VPORT]) != CAIRO_STATUS_SUCCESS) {
			fprintf(stderr, "Error creating the Cairo image surface for the RGB image\n");
			cairo_surface_destroy(com.surface[RGB_VPORT]);
			com.surface[RGB_VPORT] = NULL;
			return;
		}
	}
	// WARNING : this assumes that R, G and B buffers are already allocated and mapped
	// it seems ok, but one can probably imagine situations where it segfaults
	bufr = com.graybuf[RED_VPORT];
	bufg = com.graybuf[GREEN_VPORT];
	bufb = com.graybuf[BLUE_VPORT];
	if(bufr == NULL || bufg == NULL || bufb == NULL) {
		fprintf(stderr, "remaprgb: gray buffers not allocated for display\n");
		return;
	}
	dst = com.rgbbuf;	// index is j
	nbdata = gfit.rx * gfit.ry * 4;	// source images are 32-bit RGBA
	
	for (i = 0, j = 0; i < nbdata; i+=4) {
		dst[j++] = bufb[i];
		dst[j++] = bufg[i];
		dst[j++] = bufr[i];
		j++;		// alpha padding
	}

	// flush to ensure all writing to the image was done and redraw the surface
	cairo_surface_flush (com.surface[RGB_VPORT]);
	cairo_surface_mark_dirty (com.surface[RGB_VPORT]);
}

void activate_widgets_for_histeq(gboolean sensitive) {
	static GtkWidget *scalemax = NULL;
	static GtkWidget *scalemin = NULL;
	static GtkWidget *entrymin = NULL;
	static GtkWidget *entrymax = NULL;
	static GtkWidget *minmax   = NULL;
	static GtkWidget *hilo 	   = NULL;
	static GtkWidget *user 	   = NULL;
	
	if (!scalemax) {
		scalemax = lookup_widget("scalemax");
	    scalemin = lookup_widget("scalemin");
	    entrymin = lookup_widget("min_entry");
	    entrymax = lookup_widget("max_entry");
	    minmax 	 = lookup_widget("radiobutton_minmax");
	    hilo 	 = lookup_widget("radiobutton_hilo");
	    user 	 = lookup_widget("radiobutton_user");
	}
	gtk_widget_set_sensitive(scalemax, sensitive);
	gtk_widget_set_sensitive(scalemin, sensitive);
	gtk_widget_set_sensitive(entrymin, sensitive);
	gtk_widget_set_sensitive(entrymax, sensitive);
	gtk_widget_set_sensitive(minmax, sensitive);
	gtk_widget_set_sensitive(hilo, sensitive);
	gtk_widget_set_sensitive(user, sensitive);
}

int make_index_for_current_display(display_mode mode, WORD lo, WORD hi, int vport);
int make_index_for_rainbow(BYTE index[][3]);

void remap(int vport){
	// This function maps fit data with a linear LUT between lo and hi levels
	// to the buffer to be displayed; display only is modified
	guint x, y, i, j;
	double nb_pixels = 0.0;
	BYTE *dst, dst_pixel_value, *index, rainbow_index[UCHAR_MAX+1][3];;
	WORD *src, hi, lo, tmp_pixel_value;
	display_mode mode;
	color_map color;
	size_t hist_nb_bins;
	gboolean do_cut_over, inverted;
	gsl_histogram *histo = NULL;

	fprintf(stderr, "remap %d\n", vport);
	if (vport == RGB_VPORT) {
		remaprgb();
		return;
	}

	if ((single_image_is_loaded() && vport >= com.uniq->nb_layers) ||
			(sequence_is_loaded() && vport >= com.seq.nb_layers) ||
			(!single_image_is_loaded() && !sequence_is_loaded())) {
		fprintf(stderr, "vport is out of bounds or data is not loaded yet\n");
		return;
	}

	// allocate if not already done or the same size
	if (cairo_format_stride_for_width(CAIRO_FORMAT_RGB24, gfit.rx) != com.surface_stride[vport] ||
			gfit.ry != com.surface_height[vport] ||
			!com.surface[vport] || !com.graybuf[vport]) {
		guchar *oldbuf = com.graybuf[vport];
		fprintf(stderr, "Gray display buffers and surface (re-)allocation\n");
		if (gfit.rx == 0 || gfit.ry == 0) {
			fprintf(stderr, "gfit has a zero size, must not happen!\n");
			return;
		}
		com.surface_stride[vport] = cairo_format_stride_for_width(CAIRO_FORMAT_RGB24, gfit.rx);
		com.surface_height[vport] = gfit.ry;
		com.graybuf[vport] = realloc(com.graybuf[vport],
				com.surface_stride[vport] * gfit.ry * sizeof(guchar));
		if (com.graybuf[vport] == NULL) {
			fprintf(stderr, "Could not allocate memory for gray buffer %d (out of memory?)\n", vport);
			if (oldbuf) free(oldbuf);
			return;
		}
		if (com.surface[vport])
			cairo_surface_destroy(com.surface[vport]);
		com.surface[vport] = cairo_image_surface_create_for_data(com.graybuf[vport],
				CAIRO_FORMAT_RGB24, gfit.rx, gfit.ry,
				com.surface_stride[vport]);
		if (cairo_surface_status(com.surface[vport]) != CAIRO_STATUS_SUCCESS) {
			fprintf(stderr, "Error creating the Cairo image surface for vport %d\n", vport);
			cairo_surface_destroy(com.surface[vport]);
			com.surface[vport] = NULL;
			return;
		}
	}
	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers)
		mode = com.uniq->layers[com.cvport].rendering_mode;
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers)
		mode = com.seq.layers[com.cvport].rendering_mode;
	else return;	/* Should not happend, but it avoids warning during compilation */

	if (com.uniq) {
		hi = com.uniq->layers[vport].hi;
		lo = com.uniq->layers[vport].lo;
		do_cut_over = com.uniq->layers[vport].cut_over;
	}
	else {
		hi = com.seq.layers[vport].hi;
		lo = com.seq.layers[vport].lo;
		do_cut_over = com.seq.layers[vport].cut_over;
	}

	if (lo > hi) {
		// negative display
		WORD tmp = hi;
		hi = lo;
		lo = tmp;
		inverted = TRUE;
	}
	else inverted = FALSE;

	if (mode == HISTEQ_DISPLAY) {
		double hist_sum;
		compute_histo_for_gfit(1);
		histo = com.layers_hist[vport];
		hist_nb_bins = gsl_histogram_bins(histo);
		/*if (hist_nb_bins <= USHRT_MAX) {
			fprintf(stderr, "Error remapping: histogram is not the correct size\n");
			return;
		}*/
		nb_pixels = (double)(gfit.rx * gfit.ry);
		// build the remap_index
		if (!remap_index[vport])
			remap_index[vport] = malloc(USHRT_MAX+1);

		remap_index[vport][0] = 0;
		hist_sum = gsl_histogram_get(histo, 0);
		for (i=1; i < hist_nb_bins; i++) {
			hist_sum += gsl_histogram_get(histo, i);
			remap_index[vport][i] = round_to_BYTE((hist_sum / nb_pixels) * UCHAR_MAX_DOUBLE);
		}

		last_mode[vport] = mode;
		activate_widgets_for_histeq(FALSE);	
	} else {
		// for all other modes, the index can be reused
		make_index_for_current_display(mode, lo, hi, vport);
		activate_widgets_for_histeq(TRUE);
	}

	src = gfit.pdata[vport];	// index is i
	i = 0;
	/* Siril's FITS are stored bottom to top, so mapping needs to revert data order */
	dst = com.graybuf[vport];	// index is j
	j = (gfit.ry-1) * com.surface_stride[vport];

	color = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(lookup_widget("colormap_button")));
	
	make_index_for_rainbow(rainbow_index);
	index = remap_index[vport];
	/* cannot be parallelized because of i and j */
	for (y = 0; y < gfit.ry; y++) {
		for (x = 0; x < gfit.rx; x++, i++) {
			if (mode == HISTEQ_DISPLAY)	// special case, no lo & hi
				dst_pixel_value = index[src[i]];
			else if (do_cut_over && src[i] > hi)	// cut
				dst_pixel_value = 0;
			else {
				if (src[i] - lo < 0)
					tmp_pixel_value = 0;
				else
					tmp_pixel_value = src[i] - lo;
				dst_pixel_value = index[tmp_pixel_value];
			}
			if (inverted)
				dst_pixel_value = UCHAR_MAX - dst_pixel_value;

			double r, g, b;

			switch(color) {
				default:
				case NORMAL_COLOR:
					r = g = b = (double)dst_pixel_value;
					break;
				case RAINBOW_COLOR:
					r=(double)rainbow_index[(int)dst_pixel_value][0];
					g=(double)rainbow_index[(int)dst_pixel_value][1];
					b=(double)rainbow_index[(int)dst_pixel_value][2];
			}
			dst[j++] = round_to_BYTE(b);
			dst[j++] = round_to_BYTE(g);
			dst[j++] = round_to_BYTE(r);
			j++;		/* 8-bit padding for the 32-bit cairo RGB24 format */
		}
		j -= 2*com.surface_stride[vport];
	}

	// flush to ensure all writing to the image was done and redraw the surface
	cairo_surface_flush (com.surface[vport]);
	cairo_surface_mark_dirty (com.surface[vport]);

	test_and_allocate_reference_image(vport);
}

int make_index_for_current_display(display_mode mode, WORD lo, WORD hi, int vport) {
	float pente;
	int i;
	BYTE *index;

	/* initialization of data required to build the remap_index */
	switch (mode) {
		case NORMAL_DISPLAY:
			pente = UCHAR_MAX_SINGLE/(float)(hi - lo);
			break;
		case LOG_DISPLAY:
			pente = fabsf(UCHAR_MAX_SINGLE/logf(((float)(hi-lo))*0.1f));
			break;
		case SQRT_DISPLAY:
			pente = UCHAR_MAX_SINGLE/sqrtf((float)(hi-lo));
			break;
		case SQUARED_DISPLAY:
			pente = UCHAR_MAX_SINGLE/SQR((float)(hi-lo));
			break;
		case ASINH_DISPLAY:
			pente = UCHAR_MAX_SINGLE/asinhf(((float)(hi-lo))*0.001f);
			break;
		default:
			return 1;
	}
	if (mode != HISTEQ_DISPLAY && pente == last_pente[vport] && mode == last_mode[vport]) {
		fprintf(stdout, "Re-using previous remap_index\n");
		return 0;
	}
	fprintf(stdout, "Rebuilding remap_index\n");

	/************* Building the remap_index **************/
	if (!remap_index[vport]) {
		remap_index[vport] = malloc(USHRT_MAX+1);
		if (!remap_index[vport]) {
			fprintf(stderr, "allocation error in remap_index, aborting remap\n");
			return 1;
		}
	}
	index = remap_index[vport];
	for (i=0; i<=USHRT_MAX; i++) {
		switch (mode) {
			case LOG_DISPLAY:
				// ln(5.56*10^110) = 255
				if (i<10) index[i]=0;	/* avoid null and negative values */
				else index[i] = round_to_BYTE(logf((float)i/10.f)*pente); //10.f is arbitrary: good matching with ds9
				break;
			case SQRT_DISPLAY:
				// sqrt(2^16) = 2^8
				index[i] = round_to_BYTE(sqrtf((float)i)*pente);
				break;
			case SQUARED_DISPLAY:
				// pow(2^4,2) = 2^8
				index[i] = round_to_BYTE(SQR((float)i)*pente);
				break;
			case ASINH_DISPLAY:
				// asinh(2.78*10^110) = 255
				index[i] = round_to_BYTE(asinhf((float)i/1000.f)*pente);	//1000.f is arbitrary: good matching with ds9, could be asinhf(a*Q*i)/Q
				break;
			case NORMAL_DISPLAY:
				index[i] = round_to_BYTE((float)i*pente);
				break;
			default:
				return 1;
		}
		// check for maximum overflow, given that df/di > 0. Should not happend with round_to_BYTE
		if (index[i] == UCHAR_MAX)
			break;
	}
	if (i != USHRT_MAX+1) {
		/* no more computation needed, just fill with max value */
		for (++i; i <= USHRT_MAX; i++)
			index[i] = UCHAR_MAX;
	}

	last_pente[vport] = pente;
	last_mode[vport] = mode;
	return 0;
}

int make_index_for_rainbow(BYTE index[][3]){
	int i;
	double h, s, v, r, g, b;
	
	for(i = 0; i < UCHAR_MAX + 1; i++) {
		r = g = b = (double)i/255.0;
		rgb_to_hsv (r, g, b, &h, &s, &v);
		h=(300. - (double)i*(300./255.));	/* Arbitrary: we want h from 300 to 0 deg */
		s=1.; v=1.;										/* Saturation and Value are set to 100%  */
		hsv_to_rgb (h, s, v, &r, &g, &b);
		index[i][0] = round_to_BYTE(r*UCHAR_MAX_DOUBLE);
		index[i][1] = round_to_BYTE(g*UCHAR_MAX_DOUBLE);
		index[i][2] = round_to_BYTE(b*UCHAR_MAX_DOUBLE);
	}
	return 0;
}


/* vport can be -1 if the correct viewport should be tested */
void test_and_allocate_reference_image(int vport) {
	static GtkComboBox *cbbt_layers = NULL;
	if (cbbt_layers == NULL) {
		cbbt_layers = GTK_COMBO_BOX(gtk_builder_get_object(builder, "comboboxreglayer"));
	}
	if (vport == -1)
		vport = gtk_combo_box_get_active(cbbt_layers);

	if (sequence_is_loaded() && com.seq.current == com.seq.reference_image &&
			gtk_combo_box_get_active(cbbt_layers) == vport) {
		/* this is the registration layer and the reference frame,
		 * save the buffer for alignment preview */
		if (!com.refimage_regbuffer || !com.refimage_surface) {
			guchar *oldbuf = com.refimage_regbuffer;
			com.refimage_regbuffer = realloc(com.refimage_regbuffer,
					com.surface_stride[vport] * gfit.ry * sizeof(guchar));
			if (com.refimage_regbuffer == NULL) {
				fprintf(stderr, "** Could not allocate memory for the reference image buffer **\n");
				if (oldbuf) free(oldbuf);
				return;
			}

			if (com.refimage_surface)
				cairo_surface_destroy(com.refimage_surface);
			com.refimage_surface = cairo_image_surface_create_for_data(
					com.refimage_regbuffer,
					CAIRO_FORMAT_RGB24,
					gfit.rx, gfit.ry, com.surface_stride[vport]);
			if (cairo_surface_status(com.refimage_surface) != CAIRO_STATUS_SUCCESS) {
				fprintf(stderr, "Error creating the Cairo image surface for the reference image.\n");
				cairo_surface_destroy(com.refimage_surface);
				com.refimage_surface = NULL;
			} else {
				siril_log_message("Saved the reference frame buffer for alignment preview.\n");
				enable_view_reference_checkbox(TRUE);
			}
		}
		memcpy(com.refimage_regbuffer, com.graybuf[vport], com.surface_stride[vport] * gfit.ry * sizeof(guchar));
		cairo_surface_flush (com.refimage_surface);
		cairo_surface_mark_dirty (com.refimage_surface);
	}
}

void free_reference_image() {
	siril_log_message("Purging previously saved reference frame data.\n");
	if (com.refimage_regbuffer) {
		free(com.refimage_regbuffer);
		com.refimage_regbuffer = NULL;
	}
	if (com.refimage_surface) {
		cairo_surface_destroy(com.refimage_surface);
		com.refimage_surface = NULL;
	}
	enable_view_reference_checkbox(FALSE);
}

/* enables or disables the "display reference" checkbox in registration preview */
void enable_view_reference_checkbox(gboolean status) {
	static GtkToggleButton *check_display_ref = NULL;
	static GtkWidget *widget = NULL;
	if (check_display_ref == NULL) {
		check_display_ref = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "checkbutton_displayref"));
		widget = GTK_WIDGET(check_display_ref);
	}
	if (status && gtk_widget_get_sensitive(widget))
		return;	// may be already enabled but deactivated by user, don't force it again
	gtk_widget_set_sensitive(widget, status);
	gtk_toggle_button_set_active(check_display_ref, status);
}

gboolean redraw(int vport, int doremap){
	GtkWidget *widget;
	int i;

	if(vport >= MAXVPORT){
		fprintf(stderr,"redraw: vport=%d is larger than MAXVPORT %d\n",vport, MAXVPORT);
	}
	widget = com.vport[vport];

	switch(vport){
		case RED_VPORT:
		case BLUE_VPORT:
		case GREEN_VPORT:
			if (doremap == REMAP_ONLY) {
				remap(vport);
			} else if (doremap == REMAP_ALL) {
//~ #pragma omp parallel for private(i) schedule(dynamic,1)		//probably causes crashes in HESTEQ_MODE, no thread safe with GSL ?
				for (i=0; i<gfit.naxes[2]; i++) {
					remap(i);
				}
			}
			gtk_widget_queue_draw(widget);
			if(gfit.naxes[2] == 1)
				break;

		case RGB_VPORT:
			if(gfit.naxis == 3){
				if(doremap != REMAP_NONE){
					remaprgb();
				}
				widget=com.vport[RGB_VPORT];
				gtk_widget_queue_draw(widget);
			}
			break;
		default:
			fprintf(stderr,"redraw: unknown viewport number %d\n",vport);
			break;
	}
	fprintf(stdout, "end of redraw\n");
	com.drawn = FALSE;
	return FALSE;
}

void sliders_mode_set_state(sliders_mode sliders) {
	GtkToggleButton *radiobutton;	// Must not be static
	char *str[] = {"radiobutton_hilo", "radiobutton_minmax", "radiobutton_user"};
	void *func[] = {on_radiobutton_hilo_toggled, on_radiobutton_minmax_toggled, on_radiobutton_user_toggled};
	
	radiobutton = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, str[sliders]));
			
	g_signal_handlers_block_by_func(radiobutton, func[sliders], NULL);		
	gtk_toggle_button_set_active(radiobutton, TRUE);
	g_signal_handlers_unblock_by_func(radiobutton, func[sliders], NULL);
	com.sliders = sliders;
}

/* When rendering settings are chained, they need to be copied to other layers
 * when modified on the current layer. This procedure does that. It can be
 * called whenever a value has changed, or when the chaning has been enabled,
 * to synchronize all data.
 * Current view port is com.cvport, and data is stored in layer_info structs
 * Synchronized data: hi and lo cursors, cut over box, rendering mode.
 * DOES NOT REMAP/REDRAW.
 *
 * from_GUI: TRUE if get values from the GUI, FALSE if get the values from structs.
 * Returns 1 if chained, 0 if not.
 */
int copy_rendering_settings_when_chained(gboolean from_GUI) {
	static GtkToggleButton *chainedbutton = NULL;
	static GtkRange *range_lo = NULL, *range_hi = NULL;
	static GtkComboBox *modecombo = NULL;
	static GtkToggleButton *cutmax = NULL;

	gboolean is_chained;
	display_mode mode;
	WORD lo, hi;
	gboolean cut_over;
	int i, nb_layers;
	layer_info *layers = NULL;

	if (!chainedbutton) {	// init widgets
		chainedbutton = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_chain"));
		modecombo = GTK_COMBO_BOX(lookup_widget("combodisplay"));
		range_lo = GTK_RANGE(gtk_builder_get_object(builder, "scalemin"));
		range_hi = GTK_RANGE(gtk_builder_get_object(builder, "scalemax"));
		cutmax = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "checkcut_max"));
	}

	is_chained = gtk_toggle_button_get_active(chainedbutton);
	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers && com.uniq->layers) {
		layers = com.uniq->layers;
		nb_layers = com.uniq->nb_layers;
	} else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers && com.seq.layers) {
		layers = &com.seq.layers[com.cvport];
		nb_layers = com.seq.nb_layers;
	} else return 0;

	if (from_GUI) {
		int raw_mode = gtk_combo_box_get_active(modecombo);
		/* update values in the layer_info for cvport */
		layers[com.cvport].rendering_mode = raw_mode >= 0 ? raw_mode : NORMAL_DISPLAY;
		layers[com.cvport].lo = round_to_WORD(gtk_range_get_value(range_lo));
		layers[com.cvport].hi = round_to_WORD(gtk_range_get_value(range_hi));
		layers[com.cvport].cut_over = gtk_toggle_button_get_active(cutmax);
	}
	if (!is_chained) return 0;
	mode = layers[com.cvport].rendering_mode;
	lo = layers[com.cvport].lo;
	hi = layers[com.cvport].hi;
	cut_over = layers[com.cvport].cut_over;

	for (i=0; i<nb_layers; i++) {
		if (i == com.cvport) continue;
		layers[i].rendering_mode = mode;
		layers[i].lo = lo;
		layers[i].hi = hi;
		layers[i].cut_over = cut_over;
	}

	return 1;
}

/* when the cursor moves, update the value displayed in the textbox and save it
 * in the related layer_info. Does not change display until cursor is released. */
void on_minscale_changed (GtkRange *range, gpointer user_data)
{
	static GtkEntry *minentry = NULL;
	char buffer[10];
	if (minentry == NULL)
		minentry = GTK_ENTRY(gtk_builder_get_object(builder, "min_entry"));

	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers) {
		com.uniq->layers[com.cvport].lo = (int)gtk_range_get_value(range);
		snprintf(buffer, 6, "%d", com.uniq->layers[com.cvport].lo);
	}
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers) {
		com.seq.layers[com.cvport].lo = (int)gtk_range_get_value(range);
		snprintf(buffer, 6, "%d", com.seq.layers[com.cvport].lo);
	}
	else return;
	g_signal_handlers_block_by_func(minentry, on_min_entry_changed, NULL);
	gtk_entry_set_text(minentry, buffer);
	g_signal_handlers_unblock_by_func(minentry, on_min_entry_changed, NULL);
}

gboolean on_minscale_release(GtkWidget *widget, GdkEvent *event, gpointer user_data) {
	sliders_mode_set_state(USER);
	if (copy_rendering_settings_when_chained(FALSE))
		redraw(com.cvport, REMAP_ALL);
	else redraw(com.cvport, REMAP_ONLY);
	redraw_previews();
	return FALSE;
}

/* when the cursor moves, update the value displayed in the textbox and save it
 * in the related layer_info. Does not change display until cursor is released. */
void on_maxscale_changed (GtkRange *range, gpointer user_data)
{
	static GtkEntry *maxentry = NULL;
	char buffer[10];
	if (maxentry == NULL)
		maxentry = GTK_ENTRY(gtk_builder_get_object(builder, "max_entry"));

	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers) {
		com.uniq->layers[com.cvport].hi = (int)gtk_range_get_value(range);
		snprintf(buffer, 6, "%d", com.uniq->layers[com.cvport].hi);
	}
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers) {
		com.seq.layers[com.cvport].hi = (int)gtk_range_get_value(range);
		snprintf(buffer, 6, "%d", com.seq.layers[com.cvport].hi);
	}
	else return;
	g_signal_handlers_block_by_func(maxentry, on_max_entry_changed, NULL);
	gtk_entry_set_text(maxentry, buffer);
	g_signal_handlers_unblock_by_func(maxentry, on_max_entry_changed, NULL);
}

gboolean on_maxscale_release(GtkWidget *widget, GdkEvent *event, gpointer user_data) {
	sliders_mode_set_state(USER);
	if (copy_rendering_settings_when_chained(FALSE))
		redraw(com.cvport, REMAP_ALL);
	else redraw(com.cvport, REMAP_ONLY);
	redraw_previews();
	return FALSE;
}

/* a checkcut checkbox was toggled. Update the layer_info and others if chained. */
void on_checkcut_toggled(GtkToggleButton *togglebutton, gpointer user_data) {
	if (copy_rendering_settings_when_chained(TRUE)) 
		redraw(com.cvport, REMAP_ALL);
	else redraw(com.cvport, REMAP_ONLY);
	redraw_previews();
}

static int whichdial;

void opendial(void){
	GtkFileChooser *dialog = GTK_FILE_CHOOSER(gtk_builder_get_object(builder, "OpenDial"));
	if (!com.wd) return;
	gtk_file_chooser_set_current_folder(dialog, com.wd);
	
	remove_filters_dialog();
	switch (whichdial) {
		case OD_NULL:
			fprintf(stderr, "whichdial undefined, should not happen\n");
			return;
		case OD_FLAT:
		case OD_DARK:
		case OD_OFFSET:
		case OD_MAG:
		case OD_PHASE:
			gtk_file_chooser_set_action (dialog, GTK_FILE_CHOOSER_ACTION_OPEN);
			set_filters_dialog();
			break;
		case OD_CWD:
			gtk_file_chooser_set_action (dialog, GTK_FILE_CHOOSER_ACTION_SELECT_FOLDER);
			break;
		case OD_OPEN:
			gtk_file_chooser_set_action (dialog, GTK_FILE_CHOOSER_ACTION_OPEN);
			set_filters_dialog();
			break;
	}
	gtk_widget_show_all(GTK_WIDGET(dialog));
}

void on_darkfile_button_clicked (GtkButton *button,	gpointer user_data){
	whichdial=OD_DARK;
	opendial();
}

void on_cwd_btton_clicked (GtkButton *button, gpointer user_data){
	whichdial=OD_CWD;
	opendial();
}

void on_offsetfile_button_clicked (GtkButton *button, gpointer user_data){
	whichdial=OD_OFFSET;
	opendial();
}

void on_flatfile_button_clicked (GtkButton *button, gpointer user_data){
	whichdial=OD_FLAT;
	opendial();
}

void on_open1_activate (GtkMenuItem *menuitem, gpointer user_data){
	whichdial=OD_OPEN;
	opendial();
}

void on_open_ok_button1_clicked (GtkButton *button,	gpointer user_data){
	char *filename;
	static GtkWidget *opendial = NULL;
	if (opendial == NULL)
		opendial = lookup_widget("OpenDial");
	filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(opendial));
	
	if (!(filename))
		return;

	switch(whichdial){
		case OD_FLAT:
			gtk_entry_set_text(
					GTK_ENTRY(gtk_builder_get_object(builder, "flatname_entry")),
					filename);
			gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "useflat_button")),
					TRUE);
			if (sequence_is_loaded() || single_image_is_loaded())
				gtk_widget_set_sensitive(lookup_widget("prepro_button"), TRUE);
			break;

		case OD_DARK:
			gtk_entry_set_text(
					GTK_ENTRY(gtk_builder_get_object(builder, "darkname_entry")),
					filename);
			gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "usedark_button")),
					TRUE);
			if (sequence_is_loaded() || single_image_is_loaded())
				gtk_widget_set_sensitive(lookup_widget("prepro_button"), TRUE);
			break;

		case OD_OFFSET:
			gtk_entry_set_text(
					GTK_ENTRY(gtk_builder_get_object(builder, "offsetname_entry")),
					filename);
			gtk_toggle_button_set_active(
					GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "useoffset_button")),
					TRUE);
			if (sequence_is_loaded() || single_image_is_loaded())
				gtk_widget_set_sensitive(lookup_widget("prepro_button"), TRUE);
			break;

		case OD_CWD:
			if (!changedir(filename))
				writeinitfile();
			break;

		case OD_OPEN:
			set_cursor_waiting(TRUE);
			open_single_image(filename);
			set_cursor_waiting(FALSE);
			break;
		case OD_MAG:
			gtk_entry_set_text(GTK_ENTRY(gtk_builder_get_object(builder, "ffti_mag_entry")), filename);
		break;
		case OD_PHASE:
			gtk_entry_set_text(GTK_ENTRY(gtk_builder_get_object(builder, "ffti_phase_entry")), filename);	
		break;
	}

	whichdial = OD_NULL;
	gtk_widget_hide (opendial);
}

void on_formula_button_0_toggled(GtkToggleButton *button, gpointer user_data){
	com.preproformula = 0;
	writeinitfile();
}

void on_formula_button_1_toggled(GtkToggleButton *button, gpointer user_data){
	com.preproformula = 1;
	writeinitfile();
}

void set_filters_dialog(){
	int i;
	GtkFileChooser *dialog = GTK_FILE_CHOOSER(gtk_builder_get_object(builder, "OpenDial"));
	GtkFileFilter *filter_fits, *filter_graphics, *filter_Netpbm, *filter_IRIS, *filter_SER;
#ifdef HAVE_FFMS2
	GtkFileFilter *filter_AVI;
#endif
#ifdef HAVE_LIBRAW
	GtkFileFilter *filter_RAW;
#endif

	/*FITS FILES*/
	filter_fits = gtk_file_filter_new();
	gtk_file_filter_set_name (filter_fits, "FITS Files (*.fit, *.fits, *.fts)"); 
	gtk_file_filter_add_pattern (filter_fits, "*.fit");
	gtk_file_filter_add_pattern (filter_fits, "*.FIT");
	gtk_file_filter_add_pattern (filter_fits, "*.fits");
	gtk_file_filter_add_pattern (filter_fits, "*.FITS");
	gtk_file_filter_add_pattern (filter_fits, "*.fts");
	gtk_file_filter_add_pattern (filter_fits, "*.FTS");
	gtk_file_chooser_add_filter(dialog, filter_fits);
	gtk_file_chooser_set_filter(dialog, filter_fits);    
	
	if (whichdial == OD_OPEN) {
		/*GRAPHICS FILES*/
		filter_graphics = gtk_file_filter_new();
		char graphics_supported[256];
		sprintf(graphics_supported, "Graphics Files (*.bmp");
		gtk_file_filter_add_pattern (filter_graphics, "*.bmp");
		gtk_file_filter_add_pattern (filter_graphics, "*.BMP");

#ifdef HAVE_LIBJPEG
		strcat(graphics_supported,", *.jpg, *.jpeg");
		gtk_file_filter_add_pattern (filter_graphics, "*.jpg");
		gtk_file_filter_add_pattern (filter_graphics, "*.JPG");
		gtk_file_filter_add_pattern (filter_graphics, "*.jpeg");
		gtk_file_filter_add_pattern (filter_graphics, "*.JPEG");
#endif

#ifdef HAVE_LIBPNG
		strcat(graphics_supported,", *.png");
		gtk_file_filter_add_pattern (filter_graphics, "*.png");
		gtk_file_filter_add_pattern (filter_graphics, "*.PNG");
#endif

#ifdef HAVE_LIBTIFF
		strcat(graphics_supported,", *.tif, *.tiff");
		gtk_file_filter_add_pattern (filter_graphics, "*.tif");
		gtk_file_filter_add_pattern (filter_graphics, "*.TIF");
		gtk_file_filter_add_pattern (filter_graphics, "*.tiff");
		gtk_file_filter_add_pattern (filter_graphics, "*.TIFF");
#endif

		strcat(graphics_supported, ")");
		gtk_file_filter_set_name (filter_graphics, graphics_supported);
		gtk_file_chooser_add_filter(dialog, filter_graphics);
	
		/*NETPBM FILES*/
		filter_Netpbm = gtk_file_filter_new();
		gtk_file_filter_set_name (filter_Netpbm, "Netpbm Files (*.ppm, *.pnm, *.pgm)"); 
		gtk_file_filter_add_pattern (filter_Netpbm, "*.ppm");
		gtk_file_filter_add_pattern (filter_Netpbm, "*.PPM");
		gtk_file_filter_add_pattern (filter_Netpbm, "*.pnm");
		gtk_file_filter_add_pattern (filter_Netpbm, "*.PNM");
		gtk_file_filter_add_pattern (filter_Netpbm, "*.pgm");
		gtk_file_filter_add_pattern (filter_Netpbm, "*.PGM");
		gtk_file_chooser_add_filter(dialog, filter_Netpbm);
		
		/*IRIS FILES*/
		filter_IRIS = gtk_file_filter_new();
		gtk_file_filter_set_name (filter_IRIS, "IRIS PIC Files (*.pic)"); 
		gtk_file_filter_add_pattern (filter_IRIS, "*.pic");
		gtk_file_filter_add_pattern (filter_IRIS, "*.PIC");
		gtk_file_chooser_add_filter(dialog, filter_IRIS);
		
		/* SER FILES */
		filter_SER = gtk_file_filter_new();
		gtk_file_filter_set_name (filter_SER, "SER files (*.ser)"); 
		gtk_file_filter_add_pattern (filter_SER, "*.ser");
		gtk_file_filter_add_pattern (filter_SER, "*.SER");
		gtk_file_chooser_add_filter(dialog, filter_SER); 
		
#ifdef HAVE_FFMS2
		/* FILM FILES */
		int nb_film;
		
		nb_film = get_nb_film_ext_supported();
		filter_AVI = gtk_file_filter_new();
		gtk_file_filter_set_name (filter_AVI, "Film Files (*.avi, *.mpg, ...)"); 
		for (i = 0; i < nb_film; i++) {
			char filter[10];
			sprintf(filter, "*.%s", supported_film[i].extension);
			gtk_file_filter_add_pattern (filter_AVI, filter);
			gtk_file_filter_add_pattern (filter_AVI, convtoupper(filter));
		}
		gtk_file_chooser_add_filter(dialog, filter_AVI);
#endif	
		
#ifdef HAVE_LIBRAW
		/* RAW FILES */
		int nb_raw;
		
		nb_raw = get_nb_raw_supported();
		filter_RAW = gtk_file_filter_new();
		gtk_file_filter_set_name (filter_RAW, "RAW DSLR Camera Files"); 
		for (i = 0; i < nb_raw; i++) {
			char filter[6];
			sprintf(filter, "*.%s", supported_raw[i].extension);
			gtk_file_filter_add_pattern (filter_RAW, filter);
			gtk_file_filter_add_pattern (filter_RAW, convtoupper(filter));
		}
		gtk_file_chooser_add_filter(dialog, filter_RAW);
#endif		
	}
}

void remove_filters_dialog(){
	GtkFileChooser *dialog = GTK_FILE_CHOOSER(gtk_builder_get_object(builder, "OpenDial"));
	
	GSList* ifilters = gtk_file_chooser_list_filters(dialog);
	GSList* filters = ifilters;
	
	while (ifilters){
		gtk_file_chooser_remove_filter(dialog,GTK_FILE_FILTER(ifilters->data));
		ifilters = ifilters->next;
	}
	g_slist_free(filters);
}

void set_prepro_button_sensitiveness() {
	static GtkToggleButton *udark = NULL, *uoffset = NULL, *uflat = NULL;
	if (udark == NULL) {
		udark = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "usedark_button"));
		uoffset = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "useoffset_button"));
		uflat = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "useflat_button"));
	}
	gtk_widget_set_sensitive(lookup_widget("prepro_button"),
			(sequence_is_loaded() || single_image_is_loaded()) &&
			(gtk_toggle_button_get_active(udark) ||
			 gtk_toggle_button_get_active(uoffset) ||
			 gtk_toggle_button_get_active(uflat)));
}

void on_open_cancel_button1_clicked (GtkButton *button,	gpointer user_data){
	gtk_widget_hide (lookup_widget("OpenDial"));
}

void on_settings_activate (GtkMenuItem *menuitem, gpointer user_data){
	gtk_widget_show(lookup_widget("settings_window"));
}

void on_file1_activate (GtkMenuItem *menuitem, gpointer user_data){
	if (single_image_is_loaded() || sequence_is_loaded()) {
		gtk_widget_set_sensitive(lookup_widget("save1"), TRUE);
		if (gfit.header && !sequence_is_loaded())
			gtk_widget_set_sensitive(lookup_widget("menu_FITS_header"), TRUE);
		else
			gtk_widget_set_sensitive(lookup_widget("menu_FITS_header"), FALSE);
	}
	else {
		gtk_widget_set_sensitive(lookup_widget("save1"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menu_FITS_header"), FALSE);
	}
}

void on_menu_FITS_header_activate (GtkMenuItem *menuitem, gpointer user_data){
	if (single_image_is_loaded(BW_VPORT)) show_FITS_header(&gfit);
}

void on_close_settings_button_clicked (GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("settings_window"));
}

void on_focal_entry_changed(GtkEditable *editable, gpointer user_data){
	const gchar* focal_entry = gtk_entry_get_text(GTK_ENTRY(lookup_widget("focal_entry")));
	GtkWidget *label_ok=GTK_WIDGET(lookup_widget("label_ok"));

	gfit.focal_length = atof(focal_entry);	
	
	if (!gfit.focal_length || !gfit.pixel_size_x || !gfit.pixel_size_y) gtk_widget_set_visible(label_ok, FALSE);
	else gtk_widget_set_visible(label_ok, TRUE);
}

void on_pitchX_entry_changed(GtkEditable *editable, gpointer user_data){
	const gchar* pitchX_entry = gtk_entry_get_text(GTK_ENTRY(lookup_widget("pitchX_entry")));
	GtkWidget *label_ok=GTK_WIDGET(lookup_widget("label_ok"));

	gfit.pixel_size_x = atof(pitchX_entry);
	if (!gfit.focal_length || !gfit.pixel_size_x || !gfit.pixel_size_y) gtk_widget_set_visible(label_ok, FALSE);
	else gtk_widget_set_visible(label_ok, TRUE);	
}

void on_pitchY_entry_changed(GtkEditable *editable, gpointer user_data){
	const gchar* pitchY_entry = gtk_entry_get_text(GTK_ENTRY(lookup_widget("pitchY_entry")));
	GtkWidget *label_ok=GTK_WIDGET(lookup_widget("label_ok"));

	gfit.pixel_size_y = atof(pitchY_entry);
	if (!gfit.focal_length || !gfit.pixel_size_x || !gfit.pixel_size_y) gtk_widget_set_visible(label_ok, FALSE);
	else gtk_widget_set_visible(label_ok, TRUE);	
}

void clear_sampling_setting_box(){
	GtkComboBox *binning = GTK_COMBO_BOX(gtk_builder_get_object(builder, "combobinning"));
	GtkEntry* focal_entry=GTK_ENTRY(lookup_widget("focal_entry"));
	GtkEntry* pitchX_entry=GTK_ENTRY(lookup_widget("pitchX_entry"));
	GtkEntry* pitchY_entry=GTK_ENTRY(lookup_widget("pitchY_entry"));
	
	gtk_entry_set_text(focal_entry, "");
	gtk_entry_set_text(pitchX_entry, "");
	gtk_entry_set_text(pitchY_entry, "");
	gtk_combo_box_set_active(binning, 0);
}

void on_button_clear_sample_clicked(GtkButton *button, gpointer user_data){
	clear_sampling_setting_box();
}

void set_libraw_settings_menu_available(gboolean activate) {
	GtkNotebook* notebook = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebook3"));
	GtkWidget *page;
	page = gtk_notebook_get_nth_page(notebook, 1);
	if (!page) return;
	if (activate)
		gtk_widget_show(page);
	else
		gtk_widget_hide(page);
}

void on_comboBayer_pattern_changed(GtkComboBox* box, gpointer user_data) {
	com.raw_set.bayer_pattern = gtk_combo_box_get_active(box);
}

void on_comboBayer_inter_changed(GtkComboBox* box, gpointer user_data) {
	com.raw_set.bayer_inter = gtk_combo_box_get_active(box);
}

void set_GUI_RAW() {
	char* rawsuffix = com.raw;
	GtkEntry *entry_suffix = GTK_ENTRY(gtk_builder_get_object(builder, "entry3"));
	GtkComboBox *combo_raws = GTK_COMBO_BOX(gtk_builder_get_object(builder, "combocamera"));
	if (rawsuffix[0] == '\0') {
		strcpy(rawsuffix, "DNG");
	} /*else*/ {
		// select value com.raw in the list of avalable settings
		GtkTreeModel *tree_raws = gtk_combo_box_get_model(combo_raws);
		GtkTreeIter iter;
		gboolean valid;
		gint row_count = 0, toSelect = -1;
		for (valid = gtk_tree_model_get_iter_first(tree_raws, &iter);
				valid;
				valid = gtk_tree_model_iter_next (tree_raws, &iter))
		{
			gchar *str_data;
			gtk_tree_model_get (tree_raws, &iter, 0, &str_data, -1);
			if (strstr(str_data, com.raw))
				toSelect = row_count;
			g_free (str_data);
			if (toSelect != -1) break;
			row_count++;
		}
		if (toSelect != -1) {
			gtk_combo_box_set_active(combo_raws, toSelect);
		}
	}
	gtk_entry_set_text(entry_suffix, com.raw);
}

void set_GUI_CAMERA() {
	GtkComboBox *binning = GTK_COMBO_BOX(gtk_builder_get_object(builder, "combobinning"));

	if (gfit.focal_length) {
		char focal[8];
		sprintf(focal, "%g", gfit.focal_length);
		gtk_entry_set_text(GTK_ENTRY(lookup_widget("focal_entry")), focal);
	}
	if (gfit.pixel_size_x) {
		char pitchX[8];
		sprintf(pitchX, "%g", gfit.pixel_size_x);
		gtk_entry_set_text(GTK_ENTRY(lookup_widget("pitchX_entry")),pitchX);
	}
	if (gfit.pixel_size_y) {
		char pitchY[8];
		sprintf(pitchY, "%g", gfit.pixel_size_y);
		gtk_entry_set_text(GTK_ENTRY(lookup_widget("pitchY_entry")),pitchY);
	}
	
	if (!gfit.binning_x || !gfit.binning_y) {
		gtk_combo_box_set_active(binning, 0);
	}
	/* squared binning */
	else if (gfit.binning_x==gfit.binning_y)
		gtk_combo_box_set_active(binning, (gint) gfit.binning_x - 1);
	else {
		short coeff = gfit.binning_x>gfit.binning_y?gfit.binning_x/gfit.binning_y:gfit.binning_y/gfit.binning_x;
		switch (coeff) {
			case 2:
				gtk_combo_box_set_active(binning, 4);
				break;
			case 3:
				gtk_combo_box_set_active(binning, 5);
				break;
			default:
				siril_log_message("This binning is not handled yet");
		}
	}
}

void set_GUI_LIBRAW(){
	char radiobutton[20];

	/**********COLOR ADJUSTEMENT**************/
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_CFA")),com.raw_set.cfa);
	update_raw_cfa_interface();
		
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("Brightness_spinbutton")), com.raw_set.bright);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("Red_spinbutton")), com.raw_set.mul[0]);
	gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("Blue_spinbutton")), com.raw_set.mul[2]);
	
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_multipliers")),com.raw_set.auto_mul);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_blackpoint")),com.raw_set.user_black);
	
	/**************WHITE BALANCE**************/
	if (com.raw_set.use_camera_wb){
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_cam")),com.raw_set.use_camera_wb);
	}

	if (com.raw_set.use_auto_wb){
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_auto")),com.raw_set.use_auto_wb);
	}
	
	/********MATRIX INTERPOLATION**************/
	sprintf(radiobutton,"radiobutton_RAW%d", com.raw_set.user_qual);
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget(radiobutton)),TRUE);
	
	/********GAMMA CORRECTION**************/
	if (com.raw_set.gamm[0]==1.0 && com.raw_set.gamm[1]==1.0)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_gamm0")),TRUE);
	else if (com.raw_set.gamm[0]==2.222 && com.raw_set.gamm[1]==4.5)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_gamm1")),TRUE);
	else 
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_gamm2")),TRUE);
		
	/********** DEBAYER ******************/
	GtkComboBox *pattern = GTK_COMBO_BOX(lookup_widget("comboBayer_pattern"));
	GtkComboBox *inter = GTK_COMBO_BOX(lookup_widget("comboBayer_inter"));
	gtk_combo_box_set_active(pattern, com.raw_set.bayer_pattern);
	gtk_combo_box_set_active(inter, com.raw_set.bayer_inter);
}

void on_checkbutton_CFA_toggled(GtkToggleButton *button, gpointer user_data){
	GtkWidget *frame[8];
	int i;
	
	gboolean CFA_button = gtk_toggle_button_get_active(button);
	frame[0] = lookup_widget("frame4");
	frame[1] = lookup_widget("frame3");
	frame[2] = lookup_widget("frame19");
	frame[3] = lookup_widget("hbox8");
	frame[4] = lookup_widget("hbox9");
	frame[5] = lookup_widget("hbox11");
	frame[6] = lookup_widget("checkbutton_multipliers");
	frame[7] = lookup_widget("checkbutton_blackpoint");
	for (i = 0; i < 8; i++)
		gtk_widget_set_sensitive(frame[i], !CFA_button);
}

void on_checkbutton_cam_toggled(GtkButton *button, gpointer user_data){
	GtkToggleButton *auto_button = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_auto"));
	GtkToggleButton *cam_button = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_cam"));
	
	if (gtk_toggle_button_get_active(auto_button) == TRUE) {
		g_signal_handlers_block_by_func(auto_button, on_checkbutton_auto_toggled, NULL);
		gtk_toggle_button_set_active(auto_button, FALSE);
		g_signal_handlers_unblock_by_func(auto_button, on_checkbutton_auto_toggled, NULL);
		gtk_toggle_button_set_active(cam_button, TRUE);
	}	
}

void on_checkbutton_auto_toggled(GtkButton *button, gpointer user_data){
	GtkToggleButton *auto_button = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_auto"));
	GtkToggleButton *cam_button = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_cam"));
	
	if (gtk_toggle_button_get_active(cam_button) == TRUE) {
		g_signal_handlers_block_by_func(cam_button, on_checkbutton_cam_toggled, NULL);
		gtk_toggle_button_set_active(cam_button, FALSE);
		g_signal_handlers_unblock_by_func(cam_button, on_checkbutton_cam_toggled, NULL);
		gtk_toggle_button_set_active(auto_button, TRUE);
	}	
}

void update_libraw_interface() {
	/***** DEBAYER is updated with changed signal *******/
	
	/**********COLOR ADJUSTEMENT**************/
	com.raw_set.cfa = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_CFA")));
	update_raw_cfa_interface();

	com.raw_set.bright = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("Brightness_spinbutton")));
	com.raw_set.mul[0] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("Red_spinbutton")));
	com.raw_set.mul[2] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("Blue_spinbutton")));

	com.raw_set.auto_mul = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_multipliers")));
	com.raw_set.user_black = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_blackpoint")));

	/**************WHITE BALANCE**************/
	com.raw_set.use_camera_wb=(int)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_cam")));
	com.raw_set.use_auto_wb=(int)gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_auto")));

	/********MATRIX INTERPOLATION**************/
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_RAW0")))==TRUE){
		com.raw_set.user_qual=0;
	}
	else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_RAW1")))==TRUE){
		com.raw_set.user_qual=1;
	}
	else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_RAW2")))==TRUE){
		com.raw_set.user_qual=2;
	}
	else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_RAW3")))==TRUE){
		com.raw_set.user_qual=3;
	}
	/********GAMMA CORRECTION**************/
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_gamm0")))==TRUE){
		/* Linear Gamma Curve */
		com.raw_set.gamm[0]=1.0;
		com.raw_set.gamm[1]=1.0;
	}
	else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_gamm1")))==TRUE){
		/* BT.709 Gamma curve */
		com.raw_set.gamm[0]=2.222;
		com.raw_set.gamm[1]=4.5;
	}
	else {
		/* sRGB Gamma curve */
		com.raw_set.gamm[0]=2.40;
		com.raw_set.gamm[1]=12.92;
	}
	/* We write in config file */
	writeinitfile();
}

void on_checkbutton_auto_evaluate_toggled(GtkToggleButton *button, gpointer user_data){
	GtkWidget *entry = lookup_widget("entry_flat_norm");
	
	gtk_widget_set_sensitive(entry, !gtk_toggle_button_get_active(button));
}

void on_settings_window_hide (GtkWidget *widget, gpointer user_data){
	update_libraw_interface();
}

void on_settings_raw_extension_changed(GtkComboBoxText *box, gpointer user_data) {
	gchar *name = gtk_combo_box_text_get_active_text(box);
	if (name && name[0] != '\0') {
		int len = strlen(name);
		if (name[len-4] == ' ') {
			strcpy(com.raw, name+len-3);
			initialize_default_extension();
		}
	}
	g_free(name);
}

void on_combobinning_changed(GtkComboBox *box, gpointer user_data) {
	gint index = gtk_combo_box_get_active(box);
	
	switch(index){
		case 0:
		case 1:
		case 2:
		case 3:
			gfit.binning_x = gfit.binning_y = (short)index+1;
			break;
		case 4:
			gfit.binning_x=1;
			gfit.binning_x=2;
			break;
		case 5:
			gfit.binning_x=1;
			gfit.binning_y=3;
			break;
		default:
			fprintf(stderr, "Should not happen\n");
	}
}

void on_checkbutton_multipliers_toggled(GtkButton *button, gpointer user_data){
	gtk_widget_set_sensitive(lookup_widget("Red_spinbutton"),
			!(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_multipliers")))));
	gtk_widget_set_sensitive(lookup_widget("Blue_spinbutton"),
			!(gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_multipliers")))));
	if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_multipliers")))){
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("Red_spinbutton")), 1.0);
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("Blue_spinbutton")), 1.0);
	}
}

void on_exit_activate (GtkMenuItem *menuitem, gpointer user_data){
	exit(0);
}

void history_add_line(char *line) {
	if (!com.cmd_history) {
		com.cmd_hist_size = CMD_HISTORY_SIZE;
		com.cmd_history = calloc(com.cmd_hist_size, sizeof(const char*));
		com.cmd_hist_current = 0;
		com.cmd_hist_display = 0;
	}
	com.cmd_history[com.cmd_hist_current] = line;
	com.cmd_hist_current++;
	// circle at the end
	if (com.cmd_hist_current == com.cmd_hist_size)
		com.cmd_hist_current = 0;
	if (com.cmd_history[com.cmd_hist_current]) {
		free(com.cmd_history[com.cmd_hist_current]);
		com.cmd_history[com.cmd_hist_current] = NULL;
	}
	com.cmd_hist_display = com.cmd_hist_current;
}

/* handler for the single-line console */
gboolean on_command_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	const gchar *text;
	int handled = 0;
	static GtkEntry *entry = NULL;
	if (!entry) entry = GTK_ENTRY(widget);
	GtkEditable *editable = GTK_EDITABLE(entry);
	int entrylength = 0;
	
	switch (event->keyval) {
		case GDK_KEY_Return:
		case GDK_KEY_KP_Enter:
			handled = 1;
			text=gtk_entry_get_text(entry);
			history_add_line(strdup(text));
			if (!(processcommand(text)))
				gtk_entry_set_text(entry, "");		
			break;
		case GDK_KEY_Up:
			handled = 1;
			if (!com.cmd_history) break;
			if (com.cmd_hist_display > 0) {
				if (com.cmd_history[com.cmd_hist_display-1])
					--com.cmd_hist_display;
				// display previous entry
				gtk_entry_set_text(entry, com.cmd_history[com.cmd_hist_display]);
			}
			else if (com.cmd_history[com.cmd_hist_size-1]) {
				// ring back, display previous
				com.cmd_hist_display = com.cmd_hist_size-1;
				gtk_entry_set_text(entry, com.cmd_history[com.cmd_hist_display]);
			}
			entrylength = gtk_entry_get_text_length(entry);
			gtk_editable_set_position(editable, entrylength);
			break;
		case GDK_KEY_Down:
			handled = 1;
			if (!com.cmd_history) break;
			if (com.cmd_hist_display == com.cmd_hist_current) break;
			if (com.cmd_hist_display == com.cmd_hist_size-1) {
				if (com.cmd_hist_current == 0) {
					// ring forward, end
					gtk_entry_set_text(entry, "");
					com.cmd_hist_display++;
				}
				else if (com.cmd_history[0]) {
					// ring forward, display next
					com.cmd_hist_display = 0;
					gtk_entry_set_text(entry, com.cmd_history[0]);
				}
			} else {
				if (com.cmd_hist_display == com.cmd_hist_current-1) {
					// end
					gtk_entry_set_text(entry, "");
					com.cmd_hist_display++;
				}
				else if (com.cmd_history[com.cmd_hist_display+1]) {
					// display next
					gtk_entry_set_text(entry, com.cmd_history[++com.cmd_hist_display]);
				}
			}
			entrylength = gtk_entry_get_text_length(entry);
			gtk_editable_set_position(editable, entrylength);
			break;
		case GDK_KEY_Page_Up:
		case GDK_KEY_Page_Down:
			handled = 1;
			// go to first and last in history
			break;
	}
	return (handled == 1);
}

/* selection zone event management */
#define MAX_CALLBACKS_PER_EVENT 10
static selection_update_callback _registered_callbacks[MAX_CALLBACKS_PER_EVENT];
static int _nb_registered_callbacks = 0;

void register_selection_update_callback(selection_update_callback f) {
	if (_nb_registered_callbacks < MAX_CALLBACKS_PER_EVENT) {
		_registered_callbacks[_nb_registered_callbacks] = f;
		_nb_registered_callbacks++;
	}
}

void unregister_selection_update_callback(selection_update_callback f) {
	int i;
	for (i=0; i<_nb_registered_callbacks; ++i) {
		if (_registered_callbacks[i] == f) {
			_registered_callbacks[i] = _registered_callbacks[_nb_registered_callbacks];
			_registered_callbacks[_nb_registered_callbacks] = NULL;
			_nb_registered_callbacks--;
			return;
		}
	}
}

// send the events
void new_selection_zone() {
	int i;
	for (i=0; i<_nb_registered_callbacks; ++i) {
		_registered_callbacks[i]();
	}
}

void delete_selected_area() {
	memset(&com.selection, 0, sizeof(rectangle));
	new_selection_zone();
}

/* mouse callbacks */
gboolean on_drawingarea_button_press_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data){
	if (inimage((GdkEvent *)event)) {
		/* click on RGB image */
		if (widget == com.vport[RGB_VPORT]) {
			if (event->button == 3) {
				do_popup_rgbmenu(widget, event);
				return TRUE;
			}
			return FALSE;
		}

		/* else, click on gray image */
		if (event->button == 1) {	// left click
			if (mouse_status == MOUSE_ACTION_SELECT_REG_AREA) {
				if (com.drawing) {
					com.drawing = FALSE;
				} else {
					double zoom = get_zoom_val();
					com.drawing = TRUE;
					com.startX = event->x / zoom;
					com.startY = event->y / zoom;
					com.selection.h = 0;
					com.selection.w = 0;
				}
				gtk_widget_queue_draw(widget);
			}
		}
		else if (event->button == 2) {	// middle click
		
		}
	}
	return FALSE;
}

gboolean on_drawingarea_button_release_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data){
	if (inimage((GdkEvent *)event)){
		double zoom = get_zoom_val();
		gdouble zoomedX, zoomedY;
		zoomedX = event->x / zoom;
		zoomedY = event->y / zoom;
		if (event->button == 1){
			if (com.drawing && mouse_status == MOUSE_ACTION_SELECT_REG_AREA){
				com.drawing = FALSE;
				/* finalize selection rectangle coordinates */
				if (zoomedX > com.startX) {
					com.selection.x = com.startX;
					com.selection.w = zoomedX - com.selection.x;
				} else {
					com.selection.x = zoomedX;
					com.selection.w = com.startX - zoomedX;
				}
				if (zoomedY > com.startY) {
					com.selection.y = com.startY;
					com.selection.h = zoomedY - com.selection.y;
				} else {
					com.selection.y = zoomedY;
					com.selection.h = com.startY - zoomedY;
				}
				/* we have a new rectangular selection zone,
				 * or an unselection (empty zone) */
				new_selection_zone();

				/* calculate and display FWHM - not in event
				 * callbacks because it's in the same file and
				 * requires a special argument */
				calculate_fwhm(widget);
				com.drawn = TRUE;
			} else if (mouse_status == MOUSE_ACTION_SELECT_PREVIEW1) {
				set_preview_area(0, zoomedX, zoomedY);
				mouse_status = MOUSE_ACTION_SELECT_REG_AREA;
				// redraw to get the position of the new preview area
				gtk_widget_queue_draw(widget);
			} else if (mouse_status == MOUSE_ACTION_SELECT_PREVIEW2) {
				set_preview_area(1, zoomedX, zoomedY);
				mouse_status = MOUSE_ACTION_SELECT_REG_AREA;
				gtk_widget_queue_draw(widget);
			}
		}
		else if (event->button == 2){
			com.leveldrag = FALSE;
		}
		else if (event->button == 3) {
			do_popup_graymenu(widget, NULL);
//			set_preview_area(0, zoomedX, zoomedY); /* temporarly used for tests */
		}
	}
	return FALSE;
}

gboolean on_drawingarea_motion_notify_event (GtkWidget *widget,	GdkEventMotion *event, gpointer user_data){
	//static int delay = 5;
	gchar label[32] = "labeldensity";
	fits *fit = &(gfit);

	if (inimage((GdkEvent *)event)){
		if (com.leveldrag) {	// with button 2 down
			/* FIXME: CODE DISABLED
			fit->lo = (event->y * (double)fit->max[com.cvport] / 2 / (double)fit->ry);
			fit->hi = (event->x * (double)fit->max[com.cvport] * 2 / (double)fit->rx);
			--delay;
			if(delay<=0) {
				level_adjust(fit, FALSE);
				delay = 5;
			}*/
		} else {
			char buffer[45];
			char format[25], *format_base = "x: %%.%dd y: %%.%dd = %%.%dd";
			int coords_width = 3, val_width = 3;
			double zoom = get_zoom_val();
			gint zoomedX, zoomedY;
			zoomedX = (gint)(event->x / zoom);
			zoomedY = (gint)(event->y / zoom);
			if (fit->rx >= 1000 || fit->ry >= 1000)	coords_width = 4;
			if (fit->hi >= 1000) val_width = 4;
			if (fit->hi >= 10000) val_width = 5;
			sprintf(format, format_base, coords_width, coords_width, val_width);
			sprintf(buffer, format,
					zoomedX, zoomedY,
					//fit->data[(gint)(fit->rx*(gint)(fit->ry - event->y -1 )
					//	+(gint)event->x)]);
					fit->pdata[com.cvport][fit->rx*(fit->ry-zoomedY-1)+zoomedX]);
			/* TODO: fix to use the new function vport_number_to_name() */
			if (widget == com.vport[RED_VPORT])
				strcat(label, "r");
			else if (widget == com.vport[GREEN_VPORT])
				strcat(label, "g");
			else if (widget == com.vport[BLUE_VPORT])
				strcat(label, "b");
			else return FALSE;
			gtk_label_set_text(GTK_LABEL(gtk_builder_get_object(builder, label)), buffer);

			if (com.drawing){	// with button 1 down
				if (zoomedX > com.startX) {
					com.selection.x = com.startX;
					com.selection.w = zoomedX - com.selection.x;
				} else {
					com.selection.x = zoomedX;
					com.selection.w = com.startX - zoomedX;
				}

				if (zoomedY > com.startY) {
					com.selection.y = com.startY;
					com.selection.h = zoomedY - com.selection.y;
				} else {
					com.selection.y = zoomedY;
					com.selection.h = com.startY - zoomedY;
				}
				gtk_widget_queue_draw(widget);
			}
		}
	}
	return FALSE;
}

char *vport_number_to_name(int vport) {
	switch (vport) {
		case RED_VPORT:
			return strdup("red");
		case GREEN_VPORT:
			return strdup("green");
		case BLUE_VPORT:
			return strdup("blue");
		case RGB_VPORT:
			return strdup("rgb");
	}
	return NULL;
}

int match_drawing_area_widget(GtkWidget *drawing_area, gboolean allow_rgb) {
	/* could be done with a for i=0 loop, to get rid of these defines */
	if (drawing_area == com.vport[RED_VPORT])
		return RED_VPORT;
	if (drawing_area == com.vport[GREEN_VPORT])
		return GREEN_VPORT;
	else if (drawing_area == com.vport[BLUE_VPORT])
		return BLUE_VPORT;
	else if (allow_rgb && drawing_area == com.vport[RGB_VPORT])
		return RGB_VPORT;
	return -1;
}

void calculate_fwhm(GtkWidget *widget) {
	/* calculate and display FWHM */
	int layer = match_drawing_area_widget(widget, FALSE);
	if (layer != -1) {
		double fwhm_val;
		char buf[32], label_name[16];
		char *layer_name = vport_number_to_name(layer);
		GtkLabel *label;
		if (com.selection.w && com.selection.h) {	// Now we don't care about the size of the sample. Minimization checks that
			if (com.selection.w < 300 && com.selection.h < 300) {
				double roundness;
				fwhm_val = get_fwhm(&gfit, layer, &roundness);
				sprintf(buf, "fwhm = %.2f, r = %.2f", fwhm_val, roundness);
			} else strcpy(buf, "fwhm: selection is too large");
		} else {
			strcpy(buf, "fwhm: no selection");
		}
		sprintf(label_name, "labelfwhm%s", layer_name);
		free(layer_name);
		label = GTK_LABEL(gtk_builder_get_object(builder, label_name));
		gtk_label_set_text(label, buf);
	}
}

/* We give one signal event by toggle button to fix a bug. Without this solution
 * the toggled function was called 2 times 
 * one call to select new button, one call to unselect previous one */
void on_radiobutton_minmax_toggled (GtkToggleButton *togglebutton, gpointer user_data){
	if (gtk_toggle_button_get_active(togglebutton)==TRUE) {
		com.sliders = MINMAX;
		init_layers_hi_and_lo_values(com.sliders);
		set_cutoff_sliders_values();
		redraw(com.cvport, REMAP_ALL);	
		redraw_previews();
	}
}

void on_radiobutton_hilo_toggled (GtkToggleButton *togglebutton, gpointer user_data){
	if (gtk_toggle_button_get_active(togglebutton)==TRUE) {
		com.sliders = MIPSLOHI;
		init_layers_hi_and_lo_values(com.sliders);
		set_cutoff_sliders_values();
		redraw(com.cvport, REMAP_ALL);	
		redraw_previews();
	}
}

void on_radiobutton_user_toggled (GtkToggleButton *togglebutton, gpointer user_data){
	if (gtk_toggle_button_get_active(togglebutton)==TRUE) {
		com.sliders = USER;
		init_layers_hi_and_lo_values(com.sliders);
		set_cutoff_sliders_values();
		redraw(com.cvport, REMAP_ALL);	
		redraw_previews();
	}
}

void on_neg_button_clicked (GtkButton *button, gpointer user_data) {
	int tmp;
	static GtkToggleButton *chainedbutton = NULL;
	gboolean is_chained;

	set_cursor_waiting(TRUE);
	
	chainedbutton = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_chain"));
	is_chained = gtk_toggle_button_get_active(chainedbutton);

	/* swaps values of hi and lo and redraw */
	if (!is_chained) {
		if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers) {
			tmp = com.uniq->layers[com.cvport].hi;
			com.uniq->layers[com.cvport].hi = com.uniq->layers[com.cvport].lo;
			com.uniq->layers[com.cvport].lo = tmp;
		}
		else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers) {
			tmp = com.seq.layers[com.cvport].hi;
			com.seq.layers[com.cvport].hi = com.seq.layers[com.cvport].lo;
			com.seq.layers[com.cvport].lo = tmp;
		} else return;
		set_cutoff_sliders_values();
		redraw(com.cvport, REMAP_ONLY);	// sliders are only set for cvport
	}
	else {
		int i;
		if (single_image_is_loaded()) {
			for (i=0; i<com.uniq->nb_layers; i++){
				tmp = com.uniq->layers[i].hi;
				com.uniq->layers[i].hi = com.uniq->layers[i].lo;
				com.uniq->layers[i].lo = tmp;
			}
		} else if (sequence_is_loaded()) {
			for (i=0; i<com.seq.nb_layers; i++) {
				tmp = com.seq.layers[i].hi;
				com.seq.layers[i].hi = com.seq.layers[i].lo;
				com.seq.layers[i].lo = tmp;
			}
		} else return;
		set_cutoff_sliders_values();
		redraw(com.cvport, REMAP_ALL);
	}
	redraw_previews();
	set_cursor_waiting(FALSE);
}

void on_colormap_button_toggled(GtkToggleButton *togglebutton, gpointer user_data) {
	set_cursor_waiting(TRUE);
	redraw(com.cvport, REMAP_ALL);	
	redraw_previews();
	set_cursor_waiting(FALSE);
}

/* Callback for the display mode change */
void on_combodisplay_changed (GtkComboBox *widget, gpointer user_data) {
	if (copy_rendering_settings_when_chained(TRUE))
		redraw(com.cvport, REMAP_ALL);
	else redraw(com.cvport, REMAP_ONLY);
	redraw_previews();
}

void on_checkchain_toggled(GtkToggleButton *togglebutton, gpointer user_data) {
	if (copy_rendering_settings_when_chained(FALSE))
		redraw(com.cvport, REMAP_ALL);
}

void on_mirrorx_button_clicked (GtkButton *button, gpointer user_data){
	set_cursor_waiting(TRUE);
	mirrorx(&gfit, TRUE);
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();
	set_cursor_waiting(FALSE);
}

void on_mirrory_button_clicked (GtkButton *button, gpointer user_data){
	set_cursor_waiting(TRUE);
	mirrory(&gfit, TRUE);
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();
	set_cursor_waiting(FALSE);
}

void on_max_entry_changed (GtkEditable *editable, gpointer user_data){
	int value = atoi(gtk_entry_get_text(GTK_ENTRY(editable)));
	sliders_mode_set_state(USER);
	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers)
		com.uniq->layers[com.cvport].hi = value;
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers)
		com.seq.layers[com.cvport].hi = value;
	else return;
	set_cutoff_sliders_values();
	if (copy_rendering_settings_when_chained(FALSE))
		redraw(com.cvport, REMAP_ALL);
	else redraw(com.cvport, REMAP_ONLY);
	redraw_previews();
}

void on_min_entry_changed (GtkEditable *editable, gpointer user_data){
	int value = atoi(gtk_entry_get_text(GTK_ENTRY(editable)));
	sliders_mode_set_state(USER);
	if (single_image_is_loaded() && com.cvport < com.uniq->nb_layers)
		com.uniq->layers[com.cvport].lo = value;
	else if (sequence_is_loaded() && com.cvport < com.seq.nb_layers)
		com.seq.layers[com.cvport].lo = value;
	else return;
	set_cutoff_sliders_values();
	if (copy_rendering_settings_when_chained(FALSE))
		redraw(com.cvport, REMAP_ALL);
	else redraw(com.cvport, REMAP_ONLY);
	redraw_previews();
}

gboolean on_main_window_key_press_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	fprintf(stdout, "main window key event\n");
	/*if (event->keyval==GDK_KEY_Escape && (com.busy & ST_BUSY)){
		com.busy|=ST_CANCEL;
		return TRUE;
	}*/
	/* FIXME: there is a propagation problem somewhere, we have to do it by hand */
	return on_drawingarea_key_press_event (widget, event, user_data);
	//return FALSE;
}

void on_about_activate (GtkMenuItem *menuitem, gpointer user_data){
	gtk_widget_show_all(lookup_widget("about_window"));
}

void on_button1_clicked (GtkButton *button,	gpointer user_data){
	gtk_widget_hide(lookup_widget("about_window"));
}

void on_button_credits_clicked (GtkButton *button, gpointer user_data){
	char path[256];
	
	snprintf(path, 255, "%s/AUTHORS", PACKAGE_DOC_DIR);	
	read_and_show_textfile(path, "Credits");		//could do something better
}

void toggle_image_selection(int image_num) {
	char msg[60];
	if (com.seq.imgparam[image_num].incl) {
		com.seq.imgparam[image_num].incl = FALSE;
		--com.seq.selnum;
		sprintf(msg, "Image %d has been unselected from sequence\n", image_num);
	}
	else{
		com.seq.imgparam[image_num].incl = TRUE;
		++com.seq.selnum;
		sprintf(msg, "Image %d has been selected from sequence\n", image_num);
	}
	siril_log_message(msg);
	sequence_list_change_selection_index(image_num);
	update_reg_interface(FALSE);
	adjust_exclude(image_num, TRUE);
	writeseqfile(&com.seq);
}

void on_excludebutton_toggled (GtkToggleButton *togglebutton, gpointer user_data){
	if (!com.seq.imgparam){
		return;
	}
	toggle_image_selection(com.seq.current);
}

/* displays the opened image file name in the layers window.
 * if a unique file is loaded, its details are used instead of any sequence data
 */
void display_filename() {
	GtkLabel *fn_label;
	int nb_layers;
	char str[64], *filename;
	str[63] = '\0';
	if (com.uniq) {	// unique image
		filename = com.uniq->filename;
		nb_layers = com.uniq->nb_layers;
	} else {	// sequence
		filename = malloc(256);
		seq_get_image_filename(&com.seq, com.seq.current, filename);
		nb_layers = com.seq.nb_layers;
	}
	fn_label = GTK_LABEL(gtk_builder_get_object(builder, "labelfilename_red"));
	snprintf(str, 63, "%s (layer 0)", basename(filename));
	gtk_label_set_text(fn_label, str);
	if (nb_layers == 3) {	//take in charge both sequence and single image
		fn_label = GTK_LABEL(gtk_builder_get_object(builder, "labelfilename_green"));
		snprintf(str, 63, "%s (layer 1)", basename(filename));
		gtk_label_set_text(fn_label, str);
		fn_label = GTK_LABEL(gtk_builder_get_object(builder, "labelfilename_blue"));
		snprintf(str, 63, "%s (layer 2)", basename(filename));
		gtk_label_set_text(fn_label, str);
	}
}

/* set available layers in the layer list of registration */
void set_layers_for_assign() {
	static GtkComboBoxText *cbbt_layers = NULL;
	int i;
	if (cbbt_layers == NULL)
		cbbt_layers = GTK_COMBO_BOX_TEXT(gtk_builder_get_object(
					builder, "cbbt_layers_assign"));
	if (!com.seq.layers)
		return;
	gtk_combo_box_text_remove_all(cbbt_layers);
	for (i = 0; i < com.seq.nb_layers; i++) {
		char layer[100];
		if (!com.seq.layers[i].name) {
			if (com.seq.nb_layers == 1) {
				com.seq.layers[i].name = strdup(predefined_layers_colors[i].name);
				com.seq.layers[i].wavelength = predefined_layers_colors[i].wavelength;
			} else if (com.seq.nb_layers == 3) {
				com.seq.layers[i].name = strdup(predefined_layers_colors[i+1].name);
				com.seq.layers[i].wavelength = predefined_layers_colors[i+1].wavelength;
			} else {
				com.seq.layers[i].name = strdup("Unassigned");
				com.seq.layers[i].wavelength = -1.0;
			}
		}
		sprintf(layer, "%d: %s", i, com.seq.layers[i].name);
		gtk_combo_box_text_append_text(cbbt_layers, layer);
	}
	gtk_combo_box_set_active(GTK_COMBO_BOX(cbbt_layers), 0);
}

void set_layers_for_registration() {
	static GtkComboBoxText *cbbt_layers = NULL;
	int i;
	int reminder;
	
	if (cbbt_layers == NULL)
		cbbt_layers = GTK_COMBO_BOX_TEXT(gtk_builder_get_object(
					builder, "comboboxreglayer"));
	reminder = gtk_combo_box_get_active(GTK_COMBO_BOX(cbbt_layers));
	gtk_combo_box_text_remove_all(cbbt_layers);
	for (i = 0; i < com.seq.nb_layers; i++) {
		char layer[100];
		if (com.seq.layers[i].name)
			sprintf(layer, "%d: %s", i, com.seq.layers[i].name);
		else 	sprintf(layer, "%d: not affected yet", i);
		if (com.seq.regparam[i]) {
			// calculate average quality for this layer
			strcat(layer, " (*)");
		}
		gtk_combo_box_text_append_text(cbbt_layers, layer);
	}
	/* First initialization */
	if (reminder == -1){
		if (com.seq.nb_layers == 3)
			gtk_combo_box_set_active(GTK_COMBO_BOX(cbbt_layers), 1);
		else gtk_combo_box_set_active(GTK_COMBO_BOX(cbbt_layers), 0);
	}
	/* Already initialized */
	else gtk_combo_box_set_active(GTK_COMBO_BOX(cbbt_layers), reminder);
}

int get_index_in_predefined_colors_for_wavelength(double wl) {
	int i;
	for (i=0; i<sizeof(predefined_layers_colors)/sizeof(layer_info); i++) {
		if (predefined_layers_colors[i].wavelength == wl)
			return i;
	}
	return -1;
}

void on_layer_assign_selected(GtkComboBox *widget, gpointer user_data) {
	static GtkEntry *entry_name = NULL, *entry_wl = NULL;
	static GtkComboBox *cbbt_colors = NULL, *cbbt_sel = NULL;
	int layer, predef_idx;
	char wl[30];
	if (entry_name == NULL) {
		entry_name = GTK_ENTRY(gtk_builder_get_object(builder, "entrycolor"));
		entry_wl = GTK_ENTRY(gtk_builder_get_object(builder, "entrywavelen"));
		cbbt_colors = GTK_COMBO_BOX(gtk_builder_get_object(builder, "cbbt_colors"));
		//cbbt_sel = GTK_COMBO_BOX(gtk_builder_get_object(builder, "comboboxreglayer"));
		cbbt_sel = widget;
	}
	layer = gtk_combo_box_get_active(cbbt_sel);
	if (layer == -1) return;

	predef_idx = get_index_in_predefined_colors_for_wavelength(com.seq.layers[layer].wavelength);
	if (predef_idx >= 0)
		gtk_combo_box_set_active(cbbt_colors, predef_idx);
	
	gtk_entry_set_text(entry_name, com.seq.layers[layer].name);
	if (com.seq.layers[layer].wavelength > 0.0)
		sprintf(wl, "%g nm", com.seq.layers[layer].wavelength);
	else sprintf(wl, "undefined");
	gtk_entry_set_text(entry_wl, wl);
}

void display_image_number(int index) {
	static GtkSpinButton *spin = NULL;
	if (!spin)
		spin = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "imagenumber_spin"));
	char text[16];
	char format[10];
	if (com.seq.fixed <= 1)
		sprintf(format, "%%d");
	else sprintf(format, "%%.%dd", com.seq.fixed);
	snprintf(text, 15, format, com.seq.imgparam[index].filenum);
	gtk_entry_set_text (GTK_ENTRY (spin), text);
}

/* Returns :	TRUE if the value has been displayed */
gboolean on_imagenumberspin_output (GtkSpinButton *spin, gpointer user_data){
	static GtkAdjustment *adjustment = NULL;
	int index, do_display;
	if (adjustment == NULL)
		adjustment = gtk_spin_button_get_adjustment (spin);
	index = (int)gtk_adjustment_get_value (adjustment);

	//fprintf(stdout, "spinchanged output: index=%d\n", index);
	if (!sequence_is_loaded()) {
		// no sequence has been loaded yet
		//fprintf(stderr, "No sequence loaded\n");
		return FALSE;
	}
	if (index > com.seq.number ||
			com.seq.current == index) {	// already done for this one
		//fprintf(stderr, "index already processed or data not init. Current: %d, idx: %d\n",
		//		com.seq.current, index);
		return TRUE;
	}
	//fprintf(stdout, "SPINCHANGED: index=%d\n", index);

	do_display = (com.seq.imgparam[index].incl || com.show_excluded);
	return !seq_load_image(&com.seq, index, &gfit, do_display);
}

/* for the spin button to be able to display number which are not the actual value of
 * the adjustment, the output callback is used to modify the way they are displayed,
 * but the input callback is also required to do the opposite operation, i.e. convert
 * real image number into the number in the sequence, which is the value of the adj. */
gboolean on_imagenumberspin_input(GtkSpinButton *spin, gdouble *new_val, gpointer p) {
	const char *imgname = gtk_entry_get_text (GTK_ENTRY (spin));
	int imgname_int = atoi(imgname);
	int i;
	if (!sequence_is_loaded()) {
		//fprintf(stderr, "No sequence loaded\n");
		return FALSE;
	}
	i = com.seq.current;
	//fprintf(stdout, "current index in input is %d (looking for %d)\n", i, imgname_int);
	if (com.seq.imgparam[i].filenum == imgname_int) {
		*new_val = (gdouble)i;
		//fprintf(stdout, "found at 0 %d\n", i);
		return TRUE;
	} else if (i > 0 && com.seq.imgparam[i-1].filenum == imgname_int) {
		*new_val = (gdouble)(i-1);
		//fprintf(stdout, "found at -1 %d\n", i-1);
		return TRUE;
	} else if (i < com.seq.number - 1 && com.seq.imgparam[i+1].filenum == imgname_int) {
		*new_val = (gdouble)(i+1);
		//fprintf(stdout, "found at +1 %d\n", i+1);
		return TRUE;
	} else {
		/* no luck with neighbours, sweep it all */
		for (i = 0; i < com.seq.number; i++) {
			if (com.seq.imgparam[i].filenum == imgname_int) {
				*new_val = (gdouble)i;
				//fprintf(stdout, "sweep found at %d\n", i);
				return TRUE;
			}
		}
	}
	return GTK_INPUT_ERROR;
}

/* Returns:	TRUE to stop other handlers from being invoked for the event.
 *		FALSE to propagate the event further. */
gboolean on_imagenumberspin_key_release_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	static GtkAdjustment *adj = NULL;
	int n;

	if (!sequence_is_loaded()) {
		//fprintf(stderr, "No sequence loaded\n");
		return TRUE;
	}
	if (!adj)
		adj = gtk_spin_button_get_adjustment( GTK_SPIN_BUTTON (widget) );
	n = (int)gtk_adjustment_get_value(adj);
	if (n > com.seq.number) {
		return TRUE;
	}
	if (event->keyval == GDK_KEY_space) {
		toggle_image_selection(n);
	}
	return FALSE;
}

void on_seqexcludeall_button_clicked (GtkButton *button, gpointer user_data){
	GtkWidget *widget = lookup_widget("confirmlabel");
	confirm = CD_EXCALL;
	gtk_label_set_text (GTK_LABEL(widget),
			" Exclude all images ?\n (this erases previous image selection\n ... and there's no undo)");
	gtk_widget_show(lookup_widget("confirm_dialog"));
}

void on_seqselectall_button_clicked (GtkButton *button,	gpointer user_data){
	GtkWidget *widget = lookup_widget("confirmlabel");
	confirm = CD_INCALL;
	gtk_label_set_text (GTK_LABEL(widget),
			" Include all images ?\n (this erases previous image selection\n ... and there's no undo)");
	gtk_widget_show(lookup_widget("confirm_dialog"));
}

/* method handling all include or all exclude from a sequence */
void sequence_setselect_all(gboolean include_all) {
	int i;

	if (!com.seq.imgparam) return;
	for(i=0; i<=com.seq.number; ++i){
		if (com.seq.imgparam[i].incl != include_all) {
			com.seq.imgparam[i].incl = include_all;
			sequence_list_change_selection_index(i);
		}
	}
	if (include_all) {
		com.seq.selnum = com.seq.number;
		siril_log_message("Selected all images from sequence\n");
	} else {
		com.seq.selnum = 0;
		siril_log_message("Unselected all images from sequence\n");
	}
	adjust_exclude(com.seq.current, TRUE);
	update_reg_interface(FALSE);
	writeseqfile(&com.seq);
}

void on_prepro_button_clicked (GtkButton *button, gpointer user_data){
	GtkToggleButton *tbutton;
	GtkEntry *entry;
	com.preprostatus = 0;
	int success = 0;
	if (!single_image_is_loaded() && !sequence_is_loaded()) return;

	if (!single_image_is_loaded() && get_thread_run()) {
		siril_log_message("Another task is already in progress, ignoring new request.\n");
		return;
	}
	
	// if dark selected
	tbutton = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "usedark_button"));
	if (gtk_toggle_button_get_active (tbutton) == TRUE){
		const char *filename;
		entry = GTK_ENTRY(gtk_builder_get_object(builder, "darkname_entry"));
		filename = gtk_entry_get_text(entry);
		if (filename[0] == '\0') {
			gtk_toggle_button_set_active(tbutton, FALSE);
		} else {
			fits *dark_fit;
			progress_bar_set_text("Opening dark image...");
			dark_fit = calloc(1, sizeof(fits));
			if (readfits(filename, dark_fit, NULL)) {
				fprintf(stderr, "NOT USING DARK, cannot open the file\n");
				free(dark_fit);
				gtk_entry_set_text(entry, "");
			}
			else {
				com.preprostatus |= USE_DARK;
				if (single_image_is_loaded())
					com.uniq->dark = dark_fit;
				else com.seq.dark = dark_fit;
			}
		}
	}

	// if flat selected
	tbutton = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "useflat_button"));
	if (gtk_toggle_button_get_active (tbutton) == TRUE){
		const char *filename;
		entry = GTK_ENTRY(gtk_builder_get_object(builder, "flatname_entry"));
		filename = gtk_entry_get_text(entry);
		if (filename[0] == '\0') {
			gtk_toggle_button_set_active(tbutton, FALSE);
		} else {
			fits *flat_fit;
			progress_bar_set_text("Opening flat image...");
			flat_fit = calloc(1, sizeof(fits));
			if (readfits(filename, flat_fit, NULL)) {
				fprintf(stderr, "NOT USING FLAT, cannot open the file\n");
				free(flat_fit);
				gtk_entry_set_text(entry, "");
			}
			else {
				com.preprostatus |= USE_FLAT;
				if (single_image_is_loaded())
					com.uniq->flat = flat_fit;
				else com.seq.flat = flat_fit;
			}
		}
	}

	// if offset selected
	tbutton = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "useoffset_button"));
	if (gtk_toggle_button_get_active (tbutton) == TRUE){
		const char *filename;
		entry = GTK_ENTRY(gtk_builder_get_object(builder, "offsetname_entry"));
		filename = gtk_entry_get_text(entry);
		if (filename[0] == '\0') {
			gtk_toggle_button_set_active(tbutton, FALSE);
		} else {
			fits *bias_fit;
			progress_bar_set_text("Opening offset image...");
			bias_fit = calloc(1, sizeof(fits));
			if (readfits(filename, bias_fit, NULL)) {
				fprintf(stderr, "NOT USING OFFSET, cannot open the file\n");
				free(bias_fit);
				gtk_entry_set_text(entry, "");
			}
			else {
				com.preprostatus |= USE_OFFSET;
				if (single_image_is_loaded())
					com.uniq->offset = bias_fit;
				else com.seq.offset = bias_fit;
			}
		}
	}
	if (com.preprostatus == 0) return;

	// set output filename (preprocessed file name prefix)
	entry = GTK_ENTRY(gtk_builder_get_object(builder, "preproseqname_entry"));
	struct preprocessing_data *args = malloc(sizeof(struct preprocessing_data));
	GtkWidget *autobutton = lookup_widget("checkbutton_auto_evaluate");
	args->autolevel = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(autobutton));
	if (args->autolevel) {
		args->normalisation = 1.0f;	// will be updated anyway
	} else {
		GtkEntry *norm_entry = GTK_ENTRY(gtk_builder_get_object(builder, "entry_flat_norm"));
		args->normalisation = atof (gtk_entry_get_text (norm_entry));
	}
	GtkWidget *CCD_formula = lookup_widget("formula_button_1");
	args->use_ccd_formula = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(CCD_formula));

	if (single_image_is_loaded()) {
		com.uniq->ppprefix = strdup(gtk_entry_get_text(entry));

		// start preprocessing
		set_cursor_waiting(TRUE);
		control_window_switch_to_tab(OUTPUT_LOGS);
		success = seqpreprocess(args) == GPOINTER_TO_INT(0);
		if (success)
			progress_bar_reset_ready();
		else progress_bar_set_percent(0.0);
		// end_sequence_prepro is also executed there by seqpreprocess
		
		free(com.uniq->ppprefix);
		com.uniq->ppprefix = NULL;
		unique_free_preprocessing_data(com.uniq);
	} else {	// sequence, executed in a background thread
		if (com.seq.ppprefix) free(com.seq.ppprefix);
		com.seq.ppprefix = strdup(gtk_entry_get_text(entry));

		// start preprocessing
		set_cursor_waiting(TRUE);
		control_window_switch_to_tab(OUTPUT_LOGS);
		start_in_new_thread(seqpreprocess, args);
	}
}

// idle function executed at the end of the sequence preprocessing
gboolean end_sequence_prepro(gpointer p) {
	struct preprocessing_data *args = (struct preprocessing_data *)p;
	fprintf(stdout, "Ending sequence prepro idle function, retval=%d\n", args->retval);
	stop_processing_thread();	// can it be done here in case there is no thread?
	set_cursor_waiting(FALSE);
	update_used_memory();
	if (!args->retval && !single_image_is_loaded()) {
		// load the new sequence
		char *ppseqname = malloc(strlen(com.seq.ppprefix) + strlen(com.seq.seqname) + 5);
		sprintf(ppseqname, "%s%s.seq", com.seq.ppprefix, com.seq.seqname);
		check_seq();
		update_sequences_list(ppseqname);
		free(ppseqname);
	}
	sequence_free_preprocessing_data(&com.seq);
	free(args);
	return FALSE;
}

void on_showexcluded_button_toggled (GtkToggleButton *togglebutton,	gpointer user_data){
	com.show_excluded = gtk_toggle_button_get_active (togglebutton);
}

void on_ref_frame_toggled (GtkToggleButton *togglebutton, gpointer user_data){
	free_reference_image();
	if((gtk_toggle_button_get_active(togglebutton) == FALSE)){
		if (com.seq.reference_image == com.seq.current)
			com.seq.reference_image = -1;
	} else {
		com.seq.reference_image = com.seq.current;
		test_and_allocate_reference_image(-1);
	}
	sequence_list_change_reference();
	adjust_sellabel();	// reference image is named in the label
}

// from registration window
void on_selprefix_entry_changed (GtkEditable *editable, gpointer user_data){
#if 0
	const char *name;

	name=gtk_entry_get_text(GTK_ENTRY(editable));
	strncpy(com.seq.seqprefix, name, MAXGENLENGTH);
#endif
}

void on_seqproc_entry_changed (GtkComboBox *widget,	gpointer user_data){
	gchar *name = gtk_combo_box_text_get_active_text(GTK_COMBO_BOX_TEXT(widget));
	gchar msg[256];
	gchar *type;
	if (name && name[0] != '\0') {
		set_cursor_waiting(TRUE);
		if (ends_with(name, ".ser")) {
			name[strlen(name)-1] = 'q';
			type = " SER";
#ifdef HAVE_FFMS2
		} else if (!check_for_film_extensions(get_filename_ext(name))) {
#else
		} else if (ends_with(name, ".avi")) {
#endif
			int len = strlen(get_filename_ext(name));
			strncpy(name+strlen(name)-len-1, ".seq", len+1);
			type = " AVI";
		} else type = "";
		snprintf(msg, 255, "Selected%s sequence %s...", type, name);
		progress_bar_set_text(msg);
		set_seq(name);
		set_cursor_waiting(FALSE);
		progress_bar_reset_ready();
	}
	g_free(name);
}

/* signal handler for the gray window layer change */
void on_notebook1_switch_page (GtkNotebook *notebook, GtkWidget *page, guint page_num, gpointer user_data){
	com.cvport = page_num;
	set_cutoff_sliders_values();	// load the previous known values for sliders
	set_display_mode();		// change the mode in the combo box if needed
	redraw(com.cvport, REMAP_ONLY);
	calculate_fwhm(com.vport[com.cvport]);
	fill_sequence_list(&com.seq, com.cvport);
}

void on_checkseqbutton_clicked (GtkButton *button, gpointer user_data){
	set_cursor_waiting(TRUE);
	progress_bar_set_text("Searching for sequences in the current working directory...");
	check_seq();
	update_sequences_list(NULL);
	progress_bar_reset_ready();
	set_cursor_waiting(FALSE);
}

void on_confirmok_clicked (GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("confirm_dialog"));
	switch (confirm){
		case CD_INCALL:
			sequence_setselect_all(TRUE);
			break;

		case CD_EXCALL:
			sequence_setselect_all(FALSE);
			break;

		case CD_NULL:
			break;
	}
	confirm = CD_NULL;
}

void on_confirmcancel_clicked (GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("confirm_dialog"));
	confirm = CD_NULL;
}

gboolean on_drawingarea_key_press_event (GtkWidget *widget, GdkEventKey *event,	gpointer user_data){
	double oldzoom;
	gboolean handled = FALSE;
	fprintf(stdout, "drawing area key event\n");
	oldzoom = com.zoom_value;
	if (event->keyval == GDK_KEY_plus || event->keyval == GDK_KEY_KP_Add) {
		if (oldzoom < 0)
			com.zoom_value = 1.0;
		else com.zoom_value = min(ZOOM_MAX, oldzoom*2.0);
		handled = TRUE;
	}
	else if (event->keyval == GDK_KEY_minus || event->keyval == GDK_KEY_KP_Subtract) {
		if (oldzoom < 0)
			com.zoom_value = 1.0;
		else com.zoom_value = max(ZOOM_MIN, oldzoom/2.0);
		handled = TRUE;
	}
	else if (event->keyval == GDK_KEY_equal || event->keyval == GDK_KEY_KP_Multiply) {
		com.zoom_value = 1.0;
		handled = TRUE;
	}
	else if (event->keyval == GDK_KEY_z) {
		com.zoom_value = -1.0;
		handled = TRUE;
	}
	//else fprintf(stdout, "No bind found for key '%x'.\n", event->keyval);
	if (com.zoom_value != oldzoom){
		fprintf(stdout, "new zoom value: %f\n", com.zoom_value);
		zoomcombo_update_display_for_zoom();
		adjust_vport_size_to_image();
		redraw(com.cvport, REMAP_NONE);
	}
	return handled;
}

void on_dialog1_OK(GtkButton *button, gpointer user_data) {
	gtk_widget_hide (lookup_widget("dialog1"));
}

struct _dialog_data {
	const char *text;
	const char *title;
	const char *icon;
};

gboolean show_dialog_idle(gpointer p) {
	static GtkLabel *label = NULL;
	struct _dialog_data *args = (struct _dialog_data *)p;
	GtkImage *image = GTK_IMAGE(GTK_WIDGET(lookup_widget("image1")));
	if (label == NULL) {
		label = GTK_LABEL(gtk_builder_get_object(builder, "labeldialog1"));
	}
	gtk_window_set_title(GTK_WINDOW(lookup_widget("dialog1")), args->title);
	gtk_image_set_from_icon_name(image, args->icon, GTK_ICON_SIZE_DIALOG);
	gtk_label_set_text(label, args->text);
	gtk_widget_show(lookup_widget("dialog1"));
	free(args);
	return FALSE;
}

/* sets text in the label and displays the dialog window 1 */
void show_dialog(const char *text, const char *title, const char *icon) {
	struct _dialog_data *args = malloc(sizeof(struct _dialog_data));
       	args->text = text;
	args->title = title;
	args->icon = icon;
	gdk_threads_add_idle(show_dialog_idle, args);
}

void show_data_dialog(char *text, char *title){
	GtkTextView *tv = GTK_TEXT_VIEW(lookup_widget("data_txt"));
	GtkTextBuffer *tbuf = gtk_text_view_get_buffer(tv);
	GtkTextIter itDebut;
	GtkTextIter itFin;
	
	PangoFontDescription *font = pango_font_description_from_string( "Mono" );
	gtk_widget_override_font( GTK_WIDGET(tv), font );
	pango_font_description_free( font );
	
	gtk_text_buffer_get_bounds (tbuf, &itDebut, &itFin);
	gtk_text_buffer_delete (tbuf, &itDebut, &itFin);
	gtk_text_buffer_set_text (tbuf, text, strlen(text));
	gtk_window_set_title(GTK_WINDOW(lookup_widget("data_dialog")),title);
	
	gtk_widget_show_all(lookup_widget("data_dialog"));
}

void on_button_data_ok_clicked (GtkButton *button, gpointer user_data){
	gtk_widget_hide(lookup_widget("data_dialog"));
}

void on_menuitemWindows_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded() || sequence_is_loaded()){
		gtk_widget_set_sensitive(lookup_widget("menuitemgray"), TRUE);
		if (isrgb(&gfit))
			gtk_widget_set_sensitive(lookup_widget("menuitemcolor"), TRUE);
		else
			gtk_widget_set_sensitive(lookup_widget("menuitemcolor"), FALSE);
	}
	else{
		gtk_widget_set_sensitive(lookup_widget("menuitemgray"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitemcolor"), FALSE);
	}
}

void on_menuitemgray_toggled(GtkCheckMenuItem *checkmenuitem, gpointer user_data) {
	if (gtk_check_menu_item_get_active(checkmenuitem))
		gtk_widget_show_all(lookup_widget("main_window"));
	else gtk_widget_hide(lookup_widget("main_window"));
}

void on_menuitemcolor_toggled(GtkCheckMenuItem *checkmenuitem, gpointer user_data) {
	if (gtk_check_menu_item_get_active(checkmenuitem))
		gtk_widget_show_all(lookup_widget("rgb_window"));
	else gtk_widget_hide(lookup_widget("rgb_window"));
}

void on_rgb_window_hide(GtkWidget *object, gpointer user_data) {
	GtkCheckMenuItem *rgbcheck =
		GTK_CHECK_MENU_ITEM(gtk_builder_get_object(builder, "menuitemcolor"));
	gtk_check_menu_item_set_active(rgbcheck, FALSE);
}

void on_gray_window_hide(GtkWidget *object, gpointer user_data) {
	GtkCheckMenuItem *graycheck =
		GTK_CHECK_MENU_ITEM(gtk_builder_get_object(builder, "menuitemgray"));
	gtk_check_menu_item_set_active(graycheck, FALSE);
}

void show_main_gray_window() {
	GtkCheckMenuItem *graycheck =
		GTK_CHECK_MENU_ITEM(gtk_builder_get_object(builder, "menuitemgray"));
	gtk_check_menu_item_set_active(graycheck, TRUE);
	gtk_widget_show_all(lookup_widget("main_window"));
}

void show_rgb_window() {
	GtkCheckMenuItem *rgbcheck =
		GTK_CHECK_MENU_ITEM(gtk_builder_get_object(builder, "menuitemcolor"));
	gtk_check_menu_item_set_active(rgbcheck, TRUE);
	gtk_widget_show_all(lookup_widget("rgb_window"));
}
void hide_gray_window() {
	gtk_widget_hide(lookup_widget("main_window"));
}

void hide_rgb_window() {
	/* unchecking the menu item is done in the window destruction callback */
	/*GtkCheckMenuItem *rgbcheck =
		GTK_CHECK_MENU_ITEM(gtk_builder_get_object(builder, "menuitemcolor"));
	gtk_check_menu_item_set_active(rgbcheck, FALSE);*/
	gtk_widget_hide(lookup_widget("rgb_window"));
}

void toggle_histogram_window_visibility(GtkButton *button, gpointer user_data) {
	GtkWidget *window = lookup_widget("histogram_window");
	set_cursor_waiting(TRUE);
	compute_histo_for_gfit(1);		// it needs to be forced in the case where operation like background extraction have been done
	if (gtk_widget_get_visible(window))
		gtk_widget_hide(window);
	else gtk_widget_show(window);
	set_cursor_waiting(FALSE);
}

int is_histogram_visible() {
	GtkWidget *window = lookup_widget("histogram_window");
	return gtk_widget_get_visible(window);
}

void set_cursor_waiting(gboolean waiting) {
	/* http://developer.gnome.org/gdk/stable/gdk-Windows.html#gdk-window-set-cursor */
	GdkCursor *cursor;
	static GdkCursor *clock = NULL;
	static GdkWindow *window = NULL;
	GList *list;
	
	clock = gdk_cursor_new(GDK_WATCH);
	list = gdk_screen_get_toplevel_windows(gdk_screen_get_default ());

	if (waiting) {
//		fprintf(stdout, "wait cursor\n");
		cursor = clock;
	} else {
//		fprintf(stdout, "normal cursor\n");
		cursor = NULL;
	}
	while(list){
		window = GDK_WINDOW (list->data);
		gdk_window_set_cursor(window, cursor);
		gdk_display_sync(gdk_window_get_display(window));
		gdk_flush();
		list = list->next;
	}
	g_free(list);
	// The loop below must not be done when the function is called from a callback
	//while (gtk_events_pending())
	//	gtk_main_iteration_do(FALSE);
}

// This function is reentrant
void start_in_new_thread(gpointer(*f)(gpointer p), gpointer p) {
	g_mutex_lock(&com.mutex);
	if (com.run_thread || com.thread != NULL) {
		fprintf(stderr, "The processing thread is busy, stop it first.\n");
		g_mutex_unlock(&com.mutex);
		return;
	}

	com.run_thread = TRUE;
	g_mutex_unlock(&com.mutex);
	com.thread = g_thread_new("processing", f, p);
}

void stop_processing_thread() {
	if (com.thread == NULL) {
		fprintf(stderr, "The processing thread is not running, cannot stop it.\n");
		return;
	}

	set_thread_run(FALSE);

	g_thread_join(com.thread);
	com.thread = NULL;
}

void set_thread_run(gboolean b) {
	g_mutex_lock(&com.mutex);
	com.run_thread = b;
	g_mutex_unlock(&com.mutex);
}

gboolean get_thread_run() {
	gboolean retval;
	g_mutex_lock(&com.mutex);
	retval = com.run_thread;
	g_mutex_unlock(&com.mutex);
	return retval;
}

void on_processes_button_cancel_clicked(GtkButton *button, gpointer user_data){
	if (com.thread != NULL)
		siril_log_color_message("Process aborted by user\n", "red");
	stop_processing_thread();
}

/* http://developer.gnome.org/gtk3/3.4/GtkProgressBar.html */
void progress_bar_set_percent(double percent) {
	static GtkProgressBar *pbar = NULL;
	if (pbar == NULL)
		pbar = GTK_PROGRESS_BAR(gtk_builder_get_object(builder, "progressbar1"));
	if (percent < 0)
		gtk_progress_bar_pulse(pbar);
	else gtk_progress_bar_set_fraction(pbar, percent);
}

void progress_bar_set_text(const char *text) {
	static GtkProgressBar *pbar = NULL;
	if (pbar == NULL)
		pbar = GTK_PROGRESS_BAR(gtk_builder_get_object(builder, "progressbar1"));
	if (!text || text[0] == '\0')
		text = "Ready.";
	gtk_progress_bar_set_text(pbar, text);
}

void progress_bar_reset_ready() {
	progress_bar_set_text(PROGRESS_TEXT_RESET);
	progress_bar_set_percent(PROGRESS_RESET);
}

struct progress_bar_idle_data {
	char *progress_bar_text;
	double progress_bar_percent;
};

static gboolean progress_bar_idle_callback(gpointer p) {
	struct progress_bar_idle_data *data = (struct progress_bar_idle_data *)p;

	if (data->progress_bar_text) {
		progress_bar_set_text(data->progress_bar_text);
		free(data->progress_bar_text);
	}
	if (data->progress_bar_percent != PROGRESS_NONE)
		progress_bar_set_percent(data->progress_bar_percent);
	free(data);
	return FALSE;	// only run once
}

// Thread-safe progress bar update.
// text can be NULL, percent can be -1 for pulsating, -2 for nothing, or between 0 and 1 for percent
void set_progress_bar_data(const char *text, double percent) {
	struct progress_bar_idle_data *data;
	g_mutex_lock(&com.mutex);
	data = malloc(sizeof(struct progress_bar_idle_data));
	data->progress_bar_text = text ? strdup(text) : NULL;
	data->progress_bar_percent = percent;
	gdk_threads_add_idle(progress_bar_idle_callback, data);
	g_mutex_unlock(&com.mutex);
}


void on_combozoom_changed(GtkComboBox *widget, gpointer user_data) {
	gint active = gtk_combo_box_get_active(widget);
	switch (active) {
		case 0:	/* 16:1 */
			com.zoom_value = 16.;
			break;
		case 1:	/* 8:1 */
			com.zoom_value = 8.;
			break;
		case 2:	/* 4:1 */
			com.zoom_value = 4.;
			break;
		case 3: /* 2:1 */
			com.zoom_value = 2.;
			break;
		case -1:
		case 4: /* 1:1 */
			com.zoom_value = 1.;
			break;
		case 5: /* 1:2 */
			com.zoom_value = .5;
			break;
		case 6: /* 1:4 */
			com.zoom_value = .25;
			break;
		case 7: /* 1:8 */
			com.zoom_value = .125;
			break;
		case 8: /* fit to window */
			com.zoom_value = -1.;
			break;
	}
	fprintf(stdout, "zoom is now %f\n", com.zoom_value);
	adjust_vport_size_to_image();
	redraw(com.cvport, REMAP_NONE);
}

/* this function calculates the "fit to window" zoom values, given the window
 * size in argument and the image size in gfit.
 * Should not be called before displaying the main gray window when using zoom to fit */
double get_zoom_val() {
	int window_width, window_height;
	double wtmp, htmp;
	static GtkWidget *scrolledwin = NULL;
	if (scrolledwin == NULL)
		scrolledwin = GTK_WIDGET(gtk_builder_get_object(builder, "scrolledwindowr"));
	if (com.zoom_value > 0.)
		return com.zoom_value;
	/* else if zoom is < 0, it means fit to window */
	window_width = gtk_widget_get_allocated_width(scrolledwin);
	window_height = gtk_widget_get_allocated_height(scrolledwin);
	if (gfit.rx == 0 || gfit.ry == 0 || window_height <= 1 || window_width <= 1) return 1.0;
	wtmp = (double)window_width / (double)gfit.rx;
	htmp = (double)window_height / (double)gfit.ry;
	//fprintf(stdout, "computed fit to window zooms: %f, %f\n", wtmp, htmp);
	return min(wtmp,htmp);
}

void zoomcombo_update_display_for_zoom() {
	static GtkComboBox *zoomcombo = NULL;
	static double indexes[] = { 16., 8., 4., 2., 1., .5, .25, .125, /*.0625, */-1. };
	int i;
	char *msg;
	
	if (zoomcombo == NULL)
		zoomcombo = GTK_COMBO_BOX(gtk_builder_get_object(builder, "combozoom"));
	for (i = 0; i < sizeof(indexes)/sizeof(double); i++) {
		if (indexes[i] == com.zoom_value) {
			g_signal_handlers_block_by_func(zoomcombo, on_combozoom_changed, NULL);
			gtk_combo_box_set_active(zoomcombo, i);
			g_signal_handlers_unblock_by_func(zoomcombo, on_combozoom_changed, NULL);
			return;
		}
	}
	msg = siril_log_message("Unknown zoom_value value, what is the current zoom?\n");
	show_dialog(msg, "Error", "gtk-dialog-error");
}

void adjust_vport_size_to_image() {
	int vport;
	// make GtkDrawingArea the same size than the image
	// http://developer.gnome.org/gtk3/3.4/GtkWidget.html#gtk-widget-set-size-request
	double zoom = get_zoom_val();
	int w, h;
	if (zoom <= 0) return;
	w = (int)(((double)gfit.rx)*zoom);
	h = (int)(((double)gfit.ry)*zoom);
	for (vport=0; vport<MAXVPORT; vport++)
		gtk_widget_set_size_request(com.vport[vport],
				w, h);
	fprintf(stdout, "set new vport size (%d, %d)\n", w, h);
}

void scrollbars_hadjustment_changed_handler(GtkAdjustment *adjustment, gpointer user_data) {
	int i;
	double value = gtk_adjustment_get_value(adjustment);
	for (i=0; i<MAXVPORT; i++) {
		if (com.hadj[i] != adjustment) {
			gtk_adjustment_set_value(com.hadj[i], value);
		}
	}
}

void scrollbars_vadjustment_changed_handler(GtkAdjustment *adjustment, gpointer user_data) {
	int i;
	double value = gtk_adjustment_get_value(adjustment);
	for (i=0; i<MAXVPORT; i++) {
		if (com.vadj[i] != adjustment) {
			gtk_adjustment_set_value(com.vadj[i], value);
		}
	}
}

void on_comboboxreglayer_changed(GtkComboBox *widget, gpointer user_data) {
	free_reference_image();
	update_stack_interface();
}

/* when a sequence is loaded, the processing (stacking) output file name is
 * modified to include the name of the sequence */
void set_output_filename_to_sequence_name() {
	static GtkEntry *output_file = NULL;
	gchar msg[256];
	if (!output_file)
		output_file = GTK_ENTRY(gtk_builder_get_object(builder, "entryresultfile"));
	if (!com.seq.seqname || *com.seq.seqname == '\0') return;
	snprintf(msg, 255, "%s%sstacked.fit", com.seq.seqname,
			ends_with(com.seq.seqname, "_") ? "" : "_");
	gtk_entry_set_text(output_file, msg);
}

/* RGB popup menu */

void do_popup_rgbmenu(GtkWidget *my_widget, GdkEventButton *event) {
	static GtkMenu *menu = NULL;
	int button, event_time;
//	fprintf(stdout, "do popup\n");

	if (!menu)  {
		menu = GTK_MENU(gtk_builder_get_object(builder, "menurgb"));
		//g_signal_connect (GTK_WIDGET(menu), "deactivate", G_CALLBACK(gtk_widget_destroy), NULL);
		gtk_menu_attach_to_widget (GTK_MENU (menu), my_widget, NULL);
	}

	if (event) {
		button = event->button;
		event_time = event->time;
	} else {
		button = 0;
		event_time = gtk_get_current_event_time ();
	}

	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
			button, event_time);
}

gboolean rgb_area_popup_menu_handler(GtkWidget *widget) {
	do_popup_rgbmenu(widget, NULL);
	return TRUE;
}

/* Gray popup menu */

void do_popup_graymenu(GtkWidget *my_widget, GdkEventButton *event) {
	static GtkMenu *menu = NULL;
	int button, event_time;

	if (!menu)  {
		menu = GTK_MENU(gtk_builder_get_object(builder, "menugray"));
		gtk_menu_attach_to_widget (GTK_MENU (menu), my_widget, NULL);
	}

	if (event) {
		button = event->button;
		event_time = event->time;
	} else {
		button = 0;
		event_time = gtk_get_current_event_time ();
	}
	if (com.selection.w && com.selection.h){
		gtk_widget_set_sensitive(lookup_widget("menu_gray_psf"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menu_gray_pick_star"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menu_gray_crop"), TRUE);
		// TODO: add the sequence processing entry
	} else {
		gtk_widget_set_sensitive(lookup_widget("menu_gray_psf"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menu_gray_pick_star"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menu_gray_crop"), FALSE);
	}

	gtk_menu_popup (GTK_MENU (menu), NULL, NULL, NULL, NULL,
			button, event_time);
}

void Set_Programm_name_in_TIFF(){
	static GtkTextView *TIFF_txt = NULL;
	GtkTextBuffer *tbuf;
	GtkTextIter itDebut, itFin;
	char Copyright[256];
		
	if (TIFF_txt == NULL)
		TIFF_txt = GTK_TEXT_VIEW(lookup_widget("Copyright_txt"));
	
	tbuf = gtk_text_view_get_buffer(TIFF_txt);
	
	sprintf(Copyright, "%s v%s", PACKAGE, VERSION);
	Copyright[0] = toupper(Copyright[0]);			// convert siril to Siril
		
	PangoFontDescription *font = pango_font_description_from_string( "Mono" );
	gtk_widget_override_font( GTK_WIDGET(TIFF_txt), font );
	pango_font_description_free(font);
	
	gtk_text_buffer_get_bounds (tbuf, &itDebut, &itFin);
	gtk_text_buffer_delete (tbuf, &itDebut, &itFin);
	gtk_text_buffer_set_text (tbuf, Copyright, strlen(Copyright));
}

static image_type whichminisave;

void minisavedial(void){
	const gchar *name;
	gchar filename[256];
#ifdef HAVE_LIBJPEG
	GtkSpinButton *qlty_spin_button = GTK_SPIN_BUTTON(lookup_widget("quality_spinbutton"));
	gint quality = gtk_spin_button_get_value_as_int(qlty_spin_button);
#endif
#ifdef HAVE_LIBTIFF
	gint bitspersamples = 16;
	GtkToggleButton *BPS_Button = GTK_TOGGLE_BUTTON(lookup_widget("radiobutton8bits"));
	
	if (gtk_toggle_button_get_active(BPS_Button) == TRUE)
		bitspersamples = 8;
#endif
	int nplanes;
	GtkEntry *entry = GTK_ENTRY(lookup_widget("savetxt"));
	
	name=gtk_entry_get_text(entry);
	if (name[0] != '\0'){
		strcpy (filename,name);
		switch (whichminisave){
			case TYPEBMP:
				strcat (filename,".bmp");
				savebmp(filename, &gfit);
				break;
#ifdef HAVE_LIBJPEG
			case TYPEJPG:
				strcat (filename,".jpg");
				savejpg(filename, &gfit, quality);
				break;
#endif
#ifdef HAVE_LIBTIFF
			case TYPETIFF:
				strcat (filename, ".tif");
				savetif(filename, &gfit, bitspersamples);
				break;
#endif
			case TYPEFITS:
				/* Check if MIPS-HI and MIPS-LO must be updated. If yes,
				 * Values are taken from the layer 0 */
				if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_update_hilo"))) == TRUE){
					if (sequence_is_loaded() && !single_image_is_loaded()) {
						gfit.hi = com.seq.layers[RLAYER].hi;
						gfit.lo = com.seq.layers[RLAYER].lo;
					}
					else {
						gfit.hi = com.uniq->layers[RLAYER].hi;
						gfit.lo = com.uniq->layers[RLAYER].lo;
					}
				}
				strcat(filename,".fit");
				savefits(filename, &gfit);
				break;
			case TYPEPNM:
				nplanes=gfit.naxes[2];
				if (nplanes == 1){
					strcat (filename,".pgm");
					savepgm(filename, &gfit);
				}
				else if (nplanes == 3){
					strcat (filename,".ppm");
					saveppm(filename, &gfit);
				}
				else
					return;			//should not happend
				break;
			default:
				// TODO: handle unknown type
				break;
			}
		gtk_widget_hide (lookup_widget("savepopup"));
		gtk_entry_set_text(entry, "");
	}
}

void on_menu_rgb_savefits_activate(GtkMenuItem *menuitem, gpointer user_data) {
	static GtkNotebook* notebookFormat = NULL;
	static GtkWidget *savepopup = NULL;
	
	if (notebookFormat == NULL) {
		notebookFormat = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebookFormat"));
		savepopup = lookup_widget("savepopup");
	}
	if (single_image_is_loaded() || sequence_is_loaded()) {
		gtk_window_set_title(GTK_WINDOW(savepopup), "Saving FITS");
		gtk_widget_show_all(savepopup);
		gtk_notebook_set_current_page(notebookFormat, 2);
		whichminisave=TYPEFITS;
	}
}

void on_menu_rgb_savetiff_activate(GtkMenuItem *menuitem, gpointer user_data) {
	static GtkNotebook* notebookFormat = NULL;
	static GtkWidget *savepopup = NULL;
	
	PangoFontDescription *font = pango_font_description_from_string( "Mono" );

	if (notebookFormat == NULL) {
		notebookFormat = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebookFormat"));
		savepopup = lookup_widget("savepopup");
	}
	
	if (single_image_is_loaded() || sequence_is_loaded()) {
		Set_Programm_name_in_TIFF();		//Write "Siril Version X.Y in Copyright_Txt
		gtk_widget_override_font(GTK_WIDGET(GTK_TEXT_VIEW(lookup_widget("Description_txt"))), font); //Font = Mono
		pango_font_description_free(font);
		gtk_window_set_title(GTK_WINDOW(savepopup),"Saving TIFF");
		gtk_widget_show_all(savepopup);
		gtk_notebook_set_current_page(notebookFormat, 0);
		whichminisave=TYPETIFF;
	}
}

void on_menu_rgb_save8ppm_activate(GtkMenuItem *menuitem, gpointer user_data) {
	static GtkNotebook* notebookFormat = NULL;
	static GtkWidget *savepopup = NULL;
	
	if (notebookFormat == NULL) {
		notebookFormat = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebookFormat"));
		savepopup = lookup_widget("savepopup");
	}
	
	if (single_image_is_loaded() || sequence_is_loaded()) {
		gtk_window_set_title(GTK_WINDOW(savepopup),"Saving Netpbm");
		gtk_widget_show_all(savepopup);
		gtk_notebook_set_current_page(notebookFormat, 3);
		whichminisave=TYPEPNM;
	}
}

void on_menu_rgb_savebmp_activate(GtkMenuItem *menuitem, gpointer user_data) {
	static GtkNotebook* notebookFormat = NULL;
	static GtkWidget *savepopup = NULL;
	
	if (notebookFormat == NULL) {
		notebookFormat = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebookFormat"));
		savepopup = lookup_widget("savepopup");
	}
	
	if (single_image_is_loaded() || sequence_is_loaded()) {
		gtk_window_set_title(GTK_WINDOW(savepopup),"Saving BMP");
		gtk_widget_show_all(savepopup);
		gtk_notebook_set_current_page(notebookFormat, 3);
		whichminisave=TYPEBMP;
	}
}

void on_menu_rgb_savejpg_activate(GtkMenuItem *menuitem, gpointer user_data) {
	static GtkNotebook* notebookFormat = NULL;
	static GtkWidget *savepopup = NULL;
	
	if (notebookFormat == NULL) {
		notebookFormat = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebookFormat"));
		savepopup = lookup_widget("savepopup");
	}
	
	if (single_image_is_loaded() || sequence_is_loaded()) {
		gtk_window_set_title(GTK_WINDOW(savepopup),"Saving JPG");
		if (sequence_is_loaded() && !single_image_is_loaded()) {
			char filename[256];
			/* set the output file name default as the current image.jpg */
			GtkEntry *entry = GTK_ENTRY(lookup_widget("savetxt"));
			seq_get_image_filename(&com.seq, com.seq.current, filename);
			gtk_entry_set_text(entry, filename);
		}
		gtk_widget_show_all(savepopup);
		gtk_notebook_set_current_page(notebookFormat, 1);
		whichminisave=TYPEJPG;
	}
}


void on_button_savepopup_clicked(GtkButton *button, gpointer user_data){
	set_cursor_waiting(TRUE);
	minisavedial();
	set_cursor_waiting(FALSE);
}

void on_button_cancelpopup_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("savepopup"));
}

void close_tab(){
	GtkNotebook* Color_Layers = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebook1"));
	GtkWidget* page;
	
	if (com.seq.nb_layers==1 || gfit.naxes[2]==1){
		page = gtk_notebook_get_nth_page(Color_Layers, GREEN_VPORT);
		gtk_widget_hide(page);
		page = gtk_notebook_get_nth_page(Color_Layers, BLUE_VPORT);
		gtk_widget_hide(page);
		page = gtk_notebook_get_nth_page(Color_Layers, RED_VPORT);
		gtk_notebook_set_tab_label_text(Color_Layers, page, "B&W layer");
	}
	else {
		page = gtk_notebook_get_nth_page(Color_Layers, RED_VPORT);
		gtk_notebook_set_tab_label_text(Color_Layers, page, "Red layer");
		page = gtk_notebook_get_nth_page(Color_Layers, GREEN_VPORT);
		gtk_widget_show(page);
		page = gtk_notebook_get_nth_page(Color_Layers, BLUE_VPORT);
		gtk_widget_show(page);
	}
}

void activate_tab(int vport) {
	GtkNotebook* notebook = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebook1"));
	gtk_notebook_set_current_page(notebook, vport);
	// com.cvport is set in the event handler for changed page
}

void control_window_switch_to_tab(main_tabs tab) {
	GtkNotebook* notebook = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebook2"));
	gtk_notebook_set_current_page(notebook, tab);
}

void on_processing_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (isrgb(&gfit) && (!sequence_is_loaded() || (sequence_is_loaded() && com.seq.current == RESULT_IMAGE))){
		gtk_widget_set_sensitive(lookup_widget("removegreen"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_satu"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitemcalibration"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menu_channel_separation"), TRUE);
	}
	else{
		gtk_widget_set_sensitive(lookup_widget("removegreen"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_satu"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitemcalibration"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menu_channel_separation"), FALSE);
	}
	if (single_image_is_loaded() || (sequence_is_loaded()))
		gtk_widget_set_sensitive(lookup_widget("menuitem_histo"), TRUE);
	else
		gtk_widget_set_sensitive(lookup_widget("menuitem_histo"), FALSE);

	
	if (single_image_is_loaded() && (!sequence_is_loaded() || (sequence_is_loaded() && com.seq.current == RESULT_IMAGE))) {
#ifdef HAVE_OPENCV
		gtk_widget_set_sensitive(lookup_widget("menuitem_resample"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation90"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation270"), TRUE);
#else
		gtk_widget_set_sensitive(lookup_widget("menuitem_resample"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation90"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation270"), FALSE);
#endif
		gtk_widget_set_sensitive(lookup_widget("menuitem_mirrorx"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_mirrory"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_bkg_extraction"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menu_unsharpmask"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_fft"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_wavelets"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menu_wavelet_separation"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_medianfilter"), TRUE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_fixbanding"), TRUE);
	}
	else {
		gtk_widget_set_sensitive(lookup_widget("menuitem_resample"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation90"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_rotation270"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_mirrorx"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_mirrory"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_bkg_extraction"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menu_unsharpmask"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_fft"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_wavelets"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menu_wavelet_separation"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_medianfilter"), FALSE);
		gtk_widget_set_sensitive(lookup_widget("menuitem_fixbanding"), FALSE);
	}
}

void on_removegreen_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded() && isrgb(&gfit))
		gtk_widget_show_all(lookup_widget("SCNR_dialog"));
}

void on_menuitem_satu_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded() && isrgb(&gfit))
		gtk_widget_show_all(lookup_widget("satu_dialog"));
}

void on_satu_cancel_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide(lookup_widget("satu_dialog"));
}

void on_satu_Apply_clicked(GtkButton *button, gpointer user_data){
	static GtkComboBox *combo_satu = NULL;
	int combo;

	if (get_thread_run()) {
		siril_log_message("Another task is already in progress, ignoring new request.\n");
		return;
	}
	
	struct enhance_saturation_data *args = malloc(sizeof(struct enhance_saturation_data));

	args->coeff = gtk_range_get_value(GTK_RANGE(gtk_builder_get_object(builder, "scale_satu")));
	gboolean preserve = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(lookup_widget("preserve_bg")));
	
	if (args->coeff == 0.0) {
		free(args);
		return;
	}
	
	if (!combo_satu)
		combo_satu = GTK_COMBO_BOX(lookup_widget("combo_saturation"));
	combo = gtk_combo_box_get_active(combo_satu);
	
	set_cursor_waiting(TRUE);
	
	switch(combo) {
		case 0:		// Pink-Red to Red-Orange 
			args->h_min = 346.0;
			args->h_max = 20.0;
		break;
		case 1:		// Orange-Brown to Yellow
			args->h_min = 21.0;
			args->h_max = 60.0;
		break;
		case 2:		// Yellow-Green to Green-Cyan
			args->h_min = 61.0;
			args->h_max =	200.0;		
		break;
		case 3:		// Cyan
			args->h_min = 170.0;
			args->h_max = 200.0;
		break;
		case 4:		// Cyan-Blue to Blue-Magenta
			args->h_min = 201.0;
			args->h_max = 280.0;
		break;
		case 5:		// Magenta to Pink
			args->h_min = 281.0;
			args->h_max = 345.0;
		break;
		default:
		case 6:		// Global
			args->h_min = 0.0;
			args->h_max = 360.0;	
	}
	
	args->fit = &gfit;
	args->preserve = preserve;
	set_cursor_waiting(TRUE);
	start_in_new_thread(enhance_saturation, args);
}

void on_menu_unsharpmask_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded())
		gtk_widget_show_all(lookup_widget("Unsharp_mask_dialog"));
}

void on_button_unsharp_OK_clicked(GtkButton *button, gpointer user_data){
	gdouble sigma = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_unsharp_sigma")));
	gdouble coeff = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_unsharp_coeff")));
	
	set_cursor_waiting(TRUE);
	unsharp(&gfit, sigma, coeff, TRUE);
	adjust_cutoff_from_updated_gfit();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();
	set_cursor_waiting(FALSE);
	gtk_widget_hide(lookup_widget("Unsharp_mask_dialog"));
}

void on_SCNR_dialog_show(GtkWidget *widget, gpointer user_data) {
	GtkComboBox *comboscnr = GTK_COMBO_BOX(gtk_builder_get_object(builder, "combo_scnr"));
	int type = gtk_combo_box_get_active(comboscnr);
	
	if (type == -1)
		gtk_combo_box_set_active(comboscnr, 0);
}

void on_SCNR_Apply_clicked(GtkButton *button, gpointer user_data){
	/* Type 0: Average Neutral protection
	 * Type 1: Maximum Neutral protection
	 */
	int type = gtk_combo_box_get_active(GTK_COMBO_BOX(gtk_builder_get_object(builder, "combo_scnr")));
	GtkToggleButton *light_button = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "preserve_light"));
	gboolean preserve = gtk_toggle_button_get_active(light_button);
	double amount = gtk_range_get_value(GTK_RANGE(gtk_builder_get_object(builder, "scale_scnr")));
	
	if (get_thread_run()) {
		siril_log_message("Another task is already in progress, ignoring new request.\n");
		return;
	}
			
	struct scnr_data *args = malloc(sizeof(struct scnr_data));
	
	args->fit = &gfit;
	args->type = type;
	args->amount = amount;
	args->preserve = preserve;
	set_cursor_waiting(TRUE);
	start_in_new_thread(scnr, args);
}

void on_SCNR_cancel_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide(lookup_widget("SCNR_dialog"));
}

void on_combo_scnr_changed(GtkComboBoxText *box, gpointer user_data) {
	int type = gtk_combo_box_get_active(GTK_COMBO_BOX(gtk_builder_get_object(builder, "combo_scnr")));
	GtkScale *scale = GTK_SCALE(lookup_widget("scale_scnr"));
	GtkLabel *label = GTK_LABEL(lookup_widget("label56"));
	
	if (type > 1)  {
		gtk_widget_set_sensitive(GTK_WIDGET(scale), TRUE);
		gtk_widget_set_sensitive(GTK_WIDGET(label), TRUE);
	}
	else {
		gtk_widget_set_sensitive(GTK_WIDGET(scale), FALSE);
		gtk_widget_set_sensitive(GTK_WIDGET(label), FALSE);
	}
}

#ifdef HAVE_OPENCV
/* Resample */
void on_menuitem_resample_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded())
		gtk_widget_show_all(lookup_widget("resample_dialog"));
}

void on_button_resample_ok_clicked(GtkButton *button, gpointer user_data){
	double sample[2];
	sample[0] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_X")));
	sample[1] = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_Y")));
	int interpolation = gtk_combo_box_get_active(GTK_COMBO_BOX(lookup_widget("combo_interpolation")));
	
	set_cursor_waiting(TRUE);
	int toX = round_to_int((sample[0]/100.0) * gfit.rx);
	int toY = round_to_int((sample[1]/100.0) * gfit.ry);
	verbose_resize_gaussian(&gfit, toX, toY, interpolation);
	update_used_memory();
	adjust_vport_size_to_image();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();		
	set_cursor_waiting(FALSE);
	gtk_widget_hide(lookup_widget("resample_dialog"));
}

void on_button_resample_close_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide(lookup_widget("resample_dialog"));
}

void on_spinbutton_resample_X_value_changed (GtkSpinButton *spinbutton, gpointer user_data) {
	GtkToggleButton *ratio = GTK_TOGGLE_BUTTON(lookup_widget("button_sample_ratio"));
	double xvalue = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_X")));
	
	if (gtk_toggle_button_get_active(ratio))
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_Y")), xvalue);
}

void on_spinbutton_resample_Y_value_changed (GtkSpinButton *spinbutton, gpointer user_data) {
	GtkToggleButton *ratio = GTK_TOGGLE_BUTTON(lookup_widget("button_sample_ratio"));
	double yvalue = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_Y")));
	
	if (gtk_toggle_button_get_active(ratio))
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_X")), yvalue);
}

void on_button_sample_ratio_toggled(GtkToggleButton *button, gpointer user_data) {
	double xvalue = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_X")));

	if (gtk_toggle_button_get_active(button))
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_resample_Y")), xvalue);
}

/* Rotation */
void on_menuitem_rotation90_activate(GtkMenuItem *menuitem, gpointer user_data) {
	static GtkToggleButton *crop_rotation = NULL;
	int cropped;
	
	if (crop_rotation == NULL) {
		crop_rotation = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_rotation_crop"));
	}
	cropped = gtk_toggle_button_get_active(crop_rotation);
	
	set_cursor_waiting(TRUE);	
	verbose_rotate_image(&gfit, 90.0, -1, cropped);		// fast rotation, no interpolation, no crop
	adjust_vport_size_to_image();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();		
	set_cursor_waiting(FALSE);
}

void on_menuitem_rotation270_activate(GtkMenuItem *menuitem, gpointer user_data) {
	static GtkToggleButton *crop_rotation = NULL;
	int cropped;
	
	if (crop_rotation == NULL) {
		crop_rotation = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_rotation_crop"));
	}
	cropped = gtk_toggle_button_get_active(crop_rotation);
	
	set_cursor_waiting(TRUE);	
	verbose_rotate_image(&gfit, 270.0, -1, cropped);		// fast rotation, no interpolation, no crop
	adjust_vport_size_to_image();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();		
	set_cursor_waiting(FALSE);
}

void on_menuitem_rotation_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded())
		gtk_widget_show_all(lookup_widget("rotation_dialog"));
}

void on_button_rotation_close_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide(lookup_widget("rotation_dialog"));
}

void on_button_rotation_ok_clicked(GtkButton *button, gpointer user_data){
	static GtkToggleButton *crop_rotation = NULL;
	double angle = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_rotation")));
	int interpolation = gtk_combo_box_get_active(GTK_COMBO_BOX(lookup_widget("combo_interpolation_rotation")));
	int cropped;
	
	if (crop_rotation == NULL) {
		crop_rotation = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_rotation_crop"));
	}
	cropped = gtk_toggle_button_get_active(crop_rotation);

	set_cursor_waiting(TRUE);	
	verbose_rotate_image(&gfit, angle, interpolation, cropped);
	update_used_memory();
	adjust_vport_size_to_image();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();		
	set_cursor_waiting(FALSE);
	gtk_widget_hide(lookup_widget("rotation_dialog"));
}

#endif

void on_menuitem_mirrorx_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded()){
		set_cursor_waiting(TRUE);
		mirrorx(&gfit, TRUE);
		redraw(com.cvport, REMAP_ALL);
		redraw_previews();
		set_cursor_waiting(FALSE);
	}
}

void on_menuitem_mirrory_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded()){
		set_cursor_waiting(TRUE);
		mirrory(&gfit, TRUE);
		redraw(com.cvport, REMAP_ALL);
		redraw_previews();
		set_cursor_waiting(FALSE);
	}
}

/**************** GUI for Background extraction *******************/

void on_menuitem_bkg_extraction_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded())
		gtk_widget_show(lookup_widget("Bkg_extract_window"));
}

void on_button_extract_clicked(GtkButton *button, gpointer user_data){
	static GtkToggleButton *imgbutton = NULL;
	
	if (imgbutton == NULL)
		imgbutton = GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_bkg_img"));
	
	if (!gtk_toggle_button_get_active(imgbutton)) {
		char *msg = siril_log_message("Background cannot be extracted from itself. Please, click on Show Image\n");
		show_dialog(msg, "Error", "gtk-dialog-error");
		return;
	}
	set_cursor_waiting(TRUE);
	extract_BKG(&wfit[0]);
	redraw(com.cvport, REMAP_NONE);
	redraw_previews();
	set_cursor_waiting(FALSE);
}

void on_button_bkg_correct_clicked(GtkButton *button, gpointer user_data){
	static GtkToggleButton *imgbutton = NULL;
	
	if (imgbutton == NULL)
		imgbutton = GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_bkg_img"));
		
	if (!gtk_toggle_button_get_active(imgbutton)) {
		char *msg = siril_log_message("Please, apply correction on the image by clicking on \"Show Image\"\n");
		show_dialog(msg, "Error", "gtk-dialog-error");
		return;
	}

	int layer;
	GtkComboBox *operation = GTK_COMBO_BOX(lookup_widget("combo_correction"));
	int correction = gtk_combo_box_get_active(operation);
	
	set_cursor_waiting(TRUE);

	switch(correction) {
		default:
		case 0:	
			for (layer=0; layer<com.uniq->nb_layers; layer++){
				if (sub_background(&gfit, &wfit[0], layer)) {
					set_cursor_waiting(FALSE);
					return;
				}
			}
			siril_log_message("Subtraction done ...\n");
			break;
			case 1:
				if (ndiv(&gfit, &wfit[0])) {
					set_cursor_waiting(FALSE);
					return;
				}
			siril_log_message("Division done ...\n");
			break;
	}

	adjust_cutoff_from_updated_gfit();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();
	set_cursor_waiting(FALSE);
}

void on_checkbutton_bkg_boxes_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
	com.grad_boxes_drawn = gtk_toggle_button_get_active(togglebutton);
	redraw(com.cvport, REMAP_NONE);
	redraw_previews();
}

void on_radiobutton_bkg_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
	memcpy(&wfit[1], &gfit, sizeof(fits));
	memcpy(&gfit, &wfit[0], sizeof(fits));
	memcpy(&wfit[0], &wfit[1], sizeof(fits));
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();
}

void on_combobox_gradient_inter_changed (GtkComboBox* box, gpointer user_data) {
	GtkNotebook* notebook = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebook_bkg"));
	
	gtk_notebook_set_current_page(notebook, gtk_combo_box_get_active(box));
}

void on_button_bkg_extract_close_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide(lookup_widget("Bkg_extract_window"));
}

void on_Bkg_extract_window_hide (GtkWidget *widget, gpointer user_data) {
	static GtkToggleButton *imgbutton = NULL, *bkgbutton = NULL;
	
	if (imgbutton == NULL) {
		imgbutton = GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_bkg_img"));
		bkgbutton = GTK_TOGGLE_BUTTON(lookup_widget("radiobutton_bkg_bkg"));
	}
	gtk_widget_set_sensitive(lookup_widget("frame_bkg_tools"), FALSE);
	gtk_widget_set_sensitive(lookup_widget("button_bkg_correct"), FALSE);
	int remap = REMAP_NONE;
	
	set_cursor_waiting(TRUE);
	if (gtk_toggle_button_get_active(bkgbutton)) {
		gtk_toggle_button_set_active(imgbutton, TRUE);
		remap=REMAP_ALL;
	}
	if (com.grad) free(com.grad);
	com.grad = NULL;
	redraw(com.cvport, remap);
	redraw_previews();
	clearfits(&wfit[0]);
	set_cursor_waiting(FALSE);
}

/**********************************************************************/

void on_menu_channel_separation_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded() && isrgb(&gfit))
		gtk_widget_show_all(lookup_widget("extract_channel_dialog"));
}

void on_menuitem_histo_activate(GtkMenuItem *menuitem, gpointer user_data) {
	GtkWidget *window = lookup_widget("histogram_window");
	if (single_image_is_loaded() || sequence_is_loaded()) {
		set_cursor_waiting(TRUE);
		compute_histo_for_gfit(1);
		gtk_widget_show(window);
		set_cursor_waiting(FALSE);
	}
}

void on_menuitemcalibration_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded() && isrgb(&gfit)) {
		initialize_calibration_interface();
		gtk_widget_show(lookup_widget("color_calibration"));
	}
}

void on_menuitemPSF_toggled(GtkCheckMenuItem *checkmenuitem, gpointer user_data) {
	if (gtk_check_menu_item_get_active(checkmenuitem))
		gtk_widget_show_all(lookup_widget("stars_list_window"));
	else gtk_widget_hide(lookup_widget("stars_list_window"));
}

void on_stars_list_window_hide(GtkWidget *object, gpointer user_data) {
	GtkCheckMenuItem *PSFcheck =
		GTK_CHECK_MENU_ITEM(gtk_builder_get_object(builder, "menuitemPSF"));
	gtk_check_menu_item_set_active(PSFcheck, FALSE);
	com.selected_star=-1;
}

void on_sum_button_clicked(GtkButton *button, gpointer user_data){
	display_PSF(com.stars);
}

gboolean on_Stars_stored_button_release_event (GtkWidget *widget, GdkEventButton *event, gpointer user_data){
	GtkTreeView *tree = GTK_TREE_VIEW(gtk_builder_get_object(builder, "Stars_stored"));
	GtkTreeSelection *selection = GTK_TREE_SELECTION(gtk_builder_get_object(builder, "treeview-selection"));
	GtkTreeModel *tree_stars = gtk_tree_view_get_model(tree);
	GtkTreeIter iter;
	GtkTreePath *path;
	int *index;

	if (event->button == 1) {
		if (com.stars){
			if (gtk_tree_model_get_iter_first(tree_stars, &iter) == FALSE) return FALSE;	//The tree is empty
			if (gtk_tree_selection_get_selected(selection, &tree_stars, &iter)){	//get selected item
				path=gtk_tree_model_get_path(tree_stars, &iter);
				index=gtk_tree_path_get_indices(path);
				if (!index) return FALSE;
				com.selected_star=index[0];
				display_status(com.selected_star);
				gtk_tree_path_free(path);
				redraw(com.cvport, REMAP_NONE);
				redraw_previews();
			}
		}
	}
	return TRUE;
}

void on_Stars_stored_key_release_event (GtkWidget *widget, GdkEventKey *event, gpointer user_data){
	if (event->keyval == GDK_KEY_Delete || event->keyval == GDK_KEY_KP_Delete || event->keyval == GDK_KEY_BackSpace){
		remove_selected_line();
	}
	move_selected_line();
}

void on_remove_button_clicked(GtkButton *button, gpointer user_data){
	remove_selected_line();
}

void on_remove_all_button_clicked(GtkButton *button, gpointer user_data){
	remove_all_lines();
}

void on_process_starfinder_button_clicked(GtkButton *button, gpointer user_data){
	int layer=RLAYER;
	
	if (!single_image_is_loaded()) {
		siril_log_color_message("The star finder only checks for stars in images loaded outside a sequence, aborted.\n", "red");
		return;
	}
	set_cursor_waiting(TRUE);
	if (gfit.naxes[2] == 3) layer = GLAYER;
	delete_selected_area();
	com.stars = peaker(&gfit, layer);
	refresh_stars_list(com.stars);
	set_cursor_waiting(FALSE);
}

void on_export_button_clicked(GtkButton *button, gpointer user_data){
	int i=0;
	if (!com.stars) return;
	FILE *f = fopen("stars.lst", "w");

	if (f == NULL) return;
	while (com.stars[i]){
		fprintf(f,"%d\t%d\t%10.6f %10.6f %10.2f %10.2f %10.2f %10.2f %3.2f %10.3e\n",
		i+1,
		com.stars[i]->layer,
		com.stars[i]->B,
		com.stars[i]->A,
		com.stars[i]->xpos,
		com.stars[i]->ypos,
		com.stars[i]->FWHMX,
		com.stars[i]->FWHMY,
		com.stars[i]->An,
		com.stars[i]->rmse);
		i++;
	}
	fclose(f);
	siril_log_message("The file stars.lst has been created.\n");
}

void on_stars_list_window_show(GtkWidget *widget, gpointer user_data){
	fill_stars_list(com.stars);
}

void on_button_stars_list_ok_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("stars_list_window"));
}

void on_extract_channel_button_close_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("extract_channel_dialog"));
}

void on_extract_channel_button_ok_clicked(GtkButton *button, gpointer user_data){
	static GtkEntry *channel_extract_entry[3] = {NULL, NULL, NULL};
	static GtkComboBox *combo_extract_channel = NULL;
	
	if (get_thread_run()) {
		siril_log_message("Another task is already in progress, ignoring new request.\n");
		return;
	}

	struct extract_channels_data *args = malloc(sizeof(struct extract_channels_data));
	
	if (combo_extract_channel == NULL) {
		combo_extract_channel = GTK_COMBO_BOX(lookup_widget("combo_extract_colors"));
		channel_extract_entry[0] = GTK_ENTRY(lookup_widget("Ch1_extract_channel_entry"));
		channel_extract_entry[1] = GTK_ENTRY(lookup_widget("Ch2_extract_channel_entry"));
		channel_extract_entry[2] = GTK_ENTRY(lookup_widget("Ch3_extract_channel_entry"));
	}
	
	args->type = gtk_combo_box_get_active(combo_extract_channel);
	args->str_type = gtk_combo_box_get_active_id (combo_extract_channel);
		
	args->channel[0] = gtk_entry_get_text(channel_extract_entry[0]);
	args->channel[1] = gtk_entry_get_text(channel_extract_entry[1]);
	args->channel[2] = gtk_entry_get_text(channel_extract_entry[2]);

	if ((args->channel[0][0] != '\0') && (args->channel[1][0] != '\0') && (args->channel[2][0] != '\0')){
		args->fit = calloc(1, sizeof(fits));
		set_cursor_waiting(TRUE);
		copyfits(&gfit, args->fit, CP_ALLOC | CP_COPYA | CP_FORMAT, 0);
		start_in_new_thread(extract_channels, args);
	}
}

/******************* POPUP GRAY MENU *******************************/

void on_menu_gray_psf_activate(GtkMenuItem *menuitem, gpointer user_data){
	char msg[512];
	Param_GAUSS *result=NULL;
	int layer = match_drawing_area_widget(com.vport[com.cvport], FALSE);
	if (layer == -1) return;
	if (!(com.selection.h && com.selection.w))
		return;
	if (com.selection.w > 300 || com.selection.h > 300){
		sprintf(msg,"Current selection is too large.\nTo determine the PSF, please make a selection around a star.\n");
		show_dialog(msg, "Warning", "gtk-dialog-warning");
		return;
	}
	result = Get_Minimisation(&gfit, layer, &com.selection);
	if (!result) return;

	sprintf(msg,"Centroid Coordinates:\n");
	sprintf(msg, "%s\t\tx0=%.2fpx\n\t\ty0=%.2fpx\n\n",msg, result->x0+com.selection.x, com.selection.y+com.selection.h-result->y0);
	sprintf(msg,"%sFull Width Half Maximum:\n", msg);
	sprintf(msg, "%s\t\tFWHMx=%.2f%s\n\t\tFWHMy=%.2f%s\n\n", msg, result->FWHMX, result->units, result->FWHMY, result->units);
	sprintf(msg, "%sAngle:\n", msg);
	sprintf(msg, "%s\t\t%0.2fdeg\n\n", msg, result->An);
	sprintf(msg,"%sBackground Value:\n", msg);
	sprintf(msg, "%s\t\tB=%.6f\n\n", msg, result->B);
	sprintf(msg,"%sMaximal Intensity:\n", msg);
	sprintf(msg, "%s\t\tA=%.6f\n\n", msg, result->A);
	sprintf(msg,"%sRelative Magnitude:\n", msg);
	sprintf(msg, "%s\t\tm=%.2f\n\n", msg, result->Mag);
	sprintf(msg,"%sRMSE:\n", msg);
	sprintf(msg, "%s\t\tRMSE=%.3e", msg, result->rmse);
	show_data_dialog(msg, "PSF Results");
	free(result);
}

void on_menu_gray_pick_star_activate(GtkMenuItem *menuitem, gpointer user_data){
	char *msg;
	int layer = match_drawing_area_widget(com.vport[com.cvport], FALSE);
	int new_index;
	Param_GAUSS *new_star;
	GtkCheckMenuItem *PSFcheck = GTK_CHECK_MENU_ITEM(gtk_builder_get_object(builder, "menuitemPSF"));
	GtkWidget *window = lookup_widget("stars_list_window");
	
	if (layer != -1) {
		if (!(com.selection.h && com.selection.w))
			return;
		if (com.selection.w > 300 || com.selection.h > 300){
			msg=siril_log_message("Current selection is too large.\nTo determine the PSF, please make a selection around a star.\n");
			show_dialog(msg, "Warning", "gtk-dialog-warning");
			return;
		}
		new_star = add_star(&gfit, layer, &new_index);
		if (new_star) {
			add_star_to_list(new_star);
			if (!(gtk_widget_get_visible(window)))		//We open the stars_list_window
				gtk_widget_show_all(window);
				gtk_check_menu_item_set_active(PSFcheck, TRUE);
		}
		else return;
	}
	redraw(com.cvport, REMAP_NONE);
}

void on_menu_gray_crop_activate(GtkMenuItem *menuitem, gpointer user_data){
	if (!single_image_is_loaded()) {
		show_dialog("For now, crop is not available for sequence.",
				"Warning", "gtk-dialog-warning");
		return;
	}
	crop(&gfit, &com.selection);
	delete_selected_area();
	adjust_cutoff_from_updated_gfit();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();
	update_used_memory();
}

void on_menu_gray_stat_activate(GtkMenuItem *menuitem, gpointer user_data){
	char msg[512];
	stats* stat=NULL;
	int layer = match_drawing_area_widget(com.vport[com.cvport], FALSE);
	if (layer != -1) {
		if (single_image_is_loaded())
			snprintf(msg, 256, "%s ", com.uniq->filename);
		else if (sequence_is_loaded())
			sprintf(msg,"Image %d/%d from the sequence %s ",com.seq.current,com.seq.number,com.seq.seqname);
		else strcpy(msg, "unknown image");
		stat = statistics(&gfit,layer,&com.selection);
		sprintf(msg, "%s(%s layer)\n\n\n", msg, stat->layername);
		if (com.selection.h && com.selection.w)
			sprintf(msg,"%sSize of selection in pixel:\t(%d,%d)\n\n",msg,com.selection.w,com.selection.h);
		sprintf(msg,"%sMean:\t\t%.1lf\n\nMedian:\t\t%.1lf\n\nSigma:\t\t%.1lf\n\n",msg,stat->mean,stat->median,stat->sigma);
		sprintf(msg,"%sMinimum:\t%.1lf\n\nMaximum:\t%.1lf",msg,stat->min,stat->max);
		show_data_dialog(msg, "Statistics");
		free(stat);
		stat=NULL;
	}
}

/************************* GUI for FFT ********************************/

void on_button_fft_apply_clicked(GtkButton *button, gpointer user_data){
	const char *mag, *phase;
	char *type, *msg, page;
	int type_order=-1;
	GtkEntry *entry_mag;
	GtkEntry *entry_phase;
	static GtkToggleButton *order = NULL;
	static GtkNotebook* notebookFFT = NULL;
	
	if (get_thread_run()) {
		siril_log_message("Another task is already in progress, ignoring new request.\n");
		return;
	}
	
	if (notebookFFT == NULL) {
		notebookFFT = GTK_NOTEBOOK(gtk_builder_get_object(builder, "notebook_fft"));
		order = GTK_TOGGLE_BUTTON(gtk_builder_get_object(builder, "fft_centered"));
	}
	
	page = gtk_notebook_get_current_page (notebookFFT);
	
	set_cursor_waiting(TRUE);
	if (page == 0) {
		if (sequence_is_loaded()) {
			msg = siril_log_message("FFt does not work with sequences\n");
			show_dialog(msg, "Error", "gtk-dialog-error");
			set_cursor_waiting(FALSE);
			return;
		}
			
		entry_mag	= GTK_ENTRY(lookup_widget("fftd_mag_entry"));
		entry_phase	= GTK_ENTRY(lookup_widget("fftd_phase_entry"));
			
		if (gtk_toggle_button_get_active (order) == TRUE)
			type_order = 0;		// centered fftd
		else type_order = 1;	// regular fftd
		type	= strdup("fftd");
		mag	= gtk_entry_get_text(entry_mag);
		phase	= gtk_entry_get_text(entry_phase);
	}
	else {
		entry_mag	= GTK_ENTRY(lookup_widget("ffti_mag_entry"));
		entry_phase	= GTK_ENTRY(lookup_widget("ffti_phase_entry"));
		type	= strdup("ffti");
		mag	= gtk_entry_get_text(entry_mag);
		phase	= gtk_entry_get_text(entry_phase);
		/* The following part avoid the segfault if the image loaded in
		 * memory does not have the same dimension than the image loaded
		 * in the inverse Fourier Transform */
		if (gfit.naxes[2] > 1){
			close_single_image();
			open_single_image(mag);
		}
	}

	if ((mag[0] != '\0') && (phase[0] != '\0')){
		struct fft_data *args = malloc(sizeof(struct fft_data));
		args->fit = &gfit;
		args->type = type;
		args->modulus = mag;
		args->phase = phase;
		args->type_order = type_order;
		set_cursor_waiting(TRUE);
		start_in_new_thread(fourier_transform, args);
	}
}

void on_button_fft_close_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("dialog_FFT"));
}

void on_menuitem_fft_activate(GtkMenuItem *menuitem, gpointer user_data){
	if (single_image_is_loaded())	
		gtk_widget_show_all (lookup_widget("dialog_FFT"));	
}

void on_ffti_mag_button_clicked(GtkButton *button, gpointer user_data){
	whichdial=OD_MAG;
	opendial();
}

void on_ffti_phase_button_clicked(GtkButton *button, gpointer user_data){
	whichdial=OD_PHASE;
	opendial();
}

void on_menuitem_medianfilter_activate(GtkMenuItem *menuitem, gpointer user_data){
	if (single_image_is_loaded())
		gtk_widget_show (lookup_widget("Median_dialog"));
}

/****** GUI for wavelets *****************/

void update_wavelets ();
void reset_scale_w ();

void on_spin_w_changed(GtkSpinButton *spinbutton, gpointer user_data){
	gtk_widget_set_sensitive(lookup_widget("button_apply_w"), TRUE);
}

void on_menuitem_wavelets_activate(GtkMenuItem *menuitem, gpointer user_data){

	if (single_image_is_loaded()) {
		reset_scale_w ();
		gtk_widget_show_all (lookup_widget("wavelets_dialog"));
	}
}

void on_wavelets_dialog_hide(GtkWidget *widget, gpointer user_data) {
	gtk_widget_set_sensitive (lookup_widget("grid_w"), FALSE);
}

void reset_scale_w () {
	static GtkSpinButton *spin_w[6] = {NULL, NULL, NULL, NULL, NULL, NULL};
	int i;
	
	if (spin_w[0] == NULL) {
		spin_w[0] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w0"));
		spin_w[1] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w1"));
		spin_w[2] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w2"));
		spin_w[3] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w3"));
		spin_w[4] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w4"));
		spin_w[5] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w5"));
	}
	
	for (i=0; i < 6; i++) {
		g_signal_handlers_block_by_func(spin_w[i], on_spin_w_changed, NULL);	
		gtk_spin_button_set_value(spin_w[i], 1.f);
		g_signal_handlers_unblock_by_func(spin_w[i], on_spin_w_changed, NULL);
	}
		
	gtk_widget_set_sensitive(lookup_widget("button_apply_w"), FALSE);
}

void on_button_apply_w_clicked (GtkButton *button, gpointer user_data){
	update_wavelets();
	gtk_widget_set_sensitive(lookup_widget("button_apply_w"), FALSE);
}

void on_button_reset_w_clicked (GtkButton *button, gpointer user_data){
	float scale[6];
	static GtkSpinButton *spin_w[6] = {NULL, NULL, NULL, NULL, NULL, NULL};
	int i;
	
	if (spin_w[0] == NULL) {
		spin_w[0] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w0"));
		spin_w[1] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w1"));
		spin_w[2] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w2"));
		spin_w[3] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w3"));
		spin_w[4] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w4"));
		spin_w[5] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w5"));
	}
	
	for (i=0; i < 6; i++)
		scale[i] = (float)gtk_spin_button_get_value(spin_w[i]);
	
	if (scale[0]==1.f && scale[1]==1.f && scale[2]==1.f && scale[3]==1.f && scale[4]==1.f && scale[5]==1.f)
		return;
	reset_scale_w ();
	update_wavelets();	
}

void on_button_ok_w_clicked (GtkButton *button, gpointer user_data){
	int need_to_be_updated = gtk_widget_get_sensitive(lookup_widget("button_apply_w"));
	if (need_to_be_updated) {
		update_wavelets();
		gtk_widget_set_sensitive(lookup_widget("button_apply_w"), FALSE);
	}
	gtk_widget_hide (lookup_widget("wavelets_dialog"));
}

void on_button_cancel_w_clicked (GtkButton *button, gpointer user_data){
	float scale[6];
	static GtkSpinButton *spin_w[6] = {NULL, NULL, NULL, NULL, NULL, NULL};
	int i;
	
	if (spin_w[0] == NULL) {
		spin_w[0] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w0"));
		spin_w[1] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w1"));
		spin_w[2] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w2"));
		spin_w[3] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w3"));
		spin_w[4] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w4"));
		spin_w[5] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w5"));
	}
		
	for (i=0; i < 6; i++)
		scale[i] = (float)gtk_spin_button_get_value(spin_w[i]);
	
	if (!(scale[0]==1.f && scale[1]==1.f && scale[2]==1.f && scale[3]==1.f && scale[4]==1.f && scale[5]==1.f)) {
		if (gtk_widget_get_sensitive(lookup_widget("grid_w")) == TRUE) {
			reset_scale_w ();
			update_wavelets();	
		}
	}
	gtk_widget_hide (lookup_widget("wavelets_dialog"));
}

void on_button_compute_w_clicked (GtkButton *button, gpointer user_data){
	int Type_Transform, Nbr_Plan, maxplan, mins, i;
	int nb_chan = gfit.naxes[2];
	float *Imag;
	char *File_Name_Transform[3] = {"r_rawdata.wave", "g_rawdata.wave", "b_rawdata.wave"}, *dir[3];
	const char *tmpdir;
	
	assert(nb_chan == 1 || nb_chan == 3);
 	
 	tmpdir = g_get_tmp_dir();

	Nbr_Plan = gtk_spin_button_get_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_plans_w")));
	Type_Transform = gtk_combo_box_get_active(GTK_COMBO_BOX(lookup_widget("combobox_type_w"))) + 1;

	mins = min (gfit.rx, gfit.ry);
	maxplan = log(mins) / log(2) - 2;

	if ( Nbr_Plan > maxplan ){
		char *msg = siril_log_message("Wavelet: maximum number of plans for this image size is %d\n", 
				maxplan);
		show_dialog(msg, "Warning", "gtk-dialog-warning");
		Nbr_Plan = maxplan;
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(lookup_widget("spinbutton_plans_w")), Nbr_Plan);
	}

	if(Type_Transform!=TO_PAVE_LINEAR && Type_Transform != TO_PAVE_BSPLINE){
		char *msg = siril_log_message("Wavelet: type must be %d or %d\n", TO_PAVE_LINEAR, TO_PAVE_BSPLINE);
		show_dialog(msg, "Warning", "gtk-dialog-warning");
	}
	
	set_cursor_waiting(TRUE);

	Imag=(float *)malloc(gfit.rx*gfit.ry*sizeof(float));

	for (i = 0; i < nb_chan; i++) {
		dir[i] = malloc(strlen(tmpdir) + strlen(File_Name_Transform[i]) + 2);
		strcpy(dir[i], tmpdir);
		strcat(dir[i], "/");
		strcat(dir[i], File_Name_Transform[i]);
		wavelet_transform_file (Imag, dir[i], Type_Transform, Nbr_Plan, gfit.pdata[i]);
		free(dir[i]);
	}
	
	free (Imag);
	Imag = NULL;
	gtk_widget_set_sensitive(lookup_widget("grid_w"), TRUE);
	set_cursor_waiting(FALSE);
	return;
}

void update_wavelets () {
	float scale[6];
	static GtkSpinButton *spin_w[6] = {NULL, NULL, NULL, NULL, NULL, NULL};
	int i;
	char *File_Name_Transform[3] = {"r_rawdata.wave", "g_rawdata.wave", "b_rawdata.wave"}, *dir[3];
	const char *tmpdir;
 	
 	tmpdir = g_get_tmp_dir();

	if (spin_w[0] == NULL) {
		spin_w[0] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w0"));
		spin_w[1] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w1"));
		spin_w[2] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w2"));
		spin_w[3] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w3"));
		spin_w[4] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w4"));
		spin_w[5] = GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "spin_w5"));
	}
	
	for (i=0; i < 6; i++)
		scale[i] = (float)gtk_spin_button_get_value(spin_w[i]);
	
	set_cursor_waiting(TRUE);
	
	for (i = 0; i < gfit.naxes[2]; i++) {
		dir[i] = malloc(strlen(tmpdir) + strlen(File_Name_Transform[i]) + 2);
		strcpy(dir[i], tmpdir);
		strcat(dir[i], "/");
		strcat(dir[i], File_Name_Transform[i]);
		wavelet_reconstruct_file (dir[i], scale, gfit.pdata[i]);
		free(dir[i]);
	}
	
	adjust_cutoff_from_updated_gfit();
	redraw(com.cvport, REMAP_ALL);
	redraw_previews();
	set_cursor_waiting(FALSE);
}

/**********************************************************************/

/****************** GUI for Wavelet Layers Extraction *****************/

void on_menu_wavelet_separation_activate(GtkMenuItem *menuitem, gpointer user_data){
	
	if (single_image_is_loaded()) {
		reset_scale_w ();
		gtk_widget_show_all (lookup_widget("extract_wavelets_layers_dialog"));		
	}
}

void on_button_extract_w_ok_clicked(GtkButton *button, gpointer user_data){
	fits *fit;
	int Nbr_Plan, Type, maxplan, mins, i;
	GtkSpinButton *Spin_Nbr_Plan = NULL;
	GtkComboBox *Combo_Wavelets_Type = NULL;
	
	if (Spin_Nbr_Plan == NULL) {
		Spin_Nbr_Plan = GTK_SPIN_BUTTON(lookup_widget("spinbutton_extract_w"));
		Combo_Wavelets_Type = GTK_COMBO_BOX(lookup_widget("combo_interpolation_extract_w"));
	}
	
	Nbr_Plan = gtk_spin_button_get_value(Spin_Nbr_Plan);
	Type = gtk_combo_box_get_active(Combo_Wavelets_Type) + 1;	// 1: linear, 2: bspline

	set_cursor_waiting(TRUE);
	mins = min (gfit.rx, gfit.ry);
	maxplan = log(mins) / log(2) - 2;

	if ( Nbr_Plan > maxplan ){
		char *msg = siril_log_message("Wavelet: maximum number of plans for this image size is %d\n", 
				maxplan);
		show_dialog(msg, "Warning", "gtk-dialog-warning");
		set_cursor_waiting(FALSE);
		return;
	}
	fit = calloc(1, sizeof(fits));
	copyfits(&gfit, fit, CP_ALLOC | CP_COPYA | CP_FORMAT, 0);

	for (i=0; i < Nbr_Plan; i++) {
		char filename[256];
		
		sprintf(filename, "layer%02d", i);
		get_wavelet_layers(fit, Nbr_Plan, i, Type);
		savefits(filename, fit);
	}
	clearfits(fit);
	update_used_memory();
	set_cursor_waiting(FALSE);
}

void on_button_extract_w_close_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("extract_wavelets_layers_dialog"));	
}

/********************** GUI for Median Filter *************************/

void on_Median_cancel_clicked(GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("Median_dialog"));
}

void on_Median_Apply_clicked(GtkButton *button, gpointer user_data){
	int combo_size = gtk_combo_box_get_active(GTK_COMBO_BOX(gtk_builder_get_object(builder, "combo_ksize_median")));
	double amount = gtk_range_get_value(GTK_RANGE(gtk_builder_get_object(builder, "scale_median")));
	int iterations = round_to_int(gtk_spin_button_get_value(GTK_SPIN_BUTTON(gtk_builder_get_object(builder, "median_button_iterations"))));
	
	if (get_thread_run()) {
		siril_log_message("Another task is already in progress, ignoring new request.\n");
		return;
	}

	struct median_filter_data *args = malloc(sizeof(struct median_filter_data));
	
	switch(combo_size) {
		default:
		case 0:
			args->ksize=3;
			break;
		case 1:
			args->ksize=5;
			break;
		case 2:
			args->ksize=7;
			break;
		case 3:
			args->ksize=9;
			break;
		case 4:
			args->ksize=11;
			break;
		case 5:
			args->ksize=13;
			break;
		case 6:
			args->ksize=15;
			break;
	}
			
	args->fit = &gfit;
	args->amount = amount;
	args->iterations = iterations;
	set_cursor_waiting(TRUE);
	start_in_new_thread(median_filter, args);

}

/**********************************************************************/

/***************** GUI for Canon Banding Reduction ********************/

void on_menuitem_fixbanding_activate(GtkMenuItem *menuitem, gpointer user_data) {
	if (single_image_is_loaded())
		gtk_widget_show (lookup_widget("canon_fixbanding_dialog"));
}

void on_button_ok_fixbanding_clicked (GtkButton *button, gpointer user_data){
	gtk_widget_hide (lookup_widget("canon_fixbanding_dialog"));
}

void on_button_apply_fixbanding_clicked (GtkButton *button, gpointer user_data) {
	static GtkRange *range_amount = NULL;
	static GtkRange *range_invsigma = NULL;
	static GtkToggleButton *toggle_protect_highlights_banding = NULL;
	double amount, invsigma;
	gboolean protect_highlights;
	
	if (get_thread_run()) {
		siril_log_message("Another task is already in progress, ignoring new request.\n");
		return;
	}
	
	struct banding_data *args = malloc(sizeof(struct banding_data));

	if (range_amount == NULL) {
		range_amount = GTK_RANGE(lookup_widget("scale_fixbanding_amount"));
		range_invsigma = GTK_RANGE(lookup_widget("scale_fixbanding_invsigma"));
		toggle_protect_highlights_banding = GTK_TOGGLE_BUTTON(lookup_widget("checkbutton_fixbanding"));
	}
	amount = gtk_range_get_value(range_amount);
	invsigma = gtk_range_get_value(range_invsigma);
	protect_highlights = gtk_toggle_button_get_active(toggle_protect_highlights_banding);

	args->fit = &gfit;
	args->protect_highlights = protect_highlights;
	args->amount = amount;
	args->sigma = invsigma;
	set_cursor_waiting(TRUE);
	start_in_new_thread(BandingEngine, args);
}

void on_checkbutton_fixbanding_toggled (GtkToggleButton *togglebutton, gpointer user_data) {
	static GtkWidget *scale_invsigma = NULL;
	gboolean is_active;
	
	if (scale_invsigma == NULL) {
		scale_invsigma = lookup_widget("scale_fixbanding_invsigma");
	}
	
	is_active = gtk_toggle_button_get_active(togglebutton);
	gtk_widget_set_sensitive(scale_invsigma, is_active);
}

/**********************************************************************/
